# Copyright 2019 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from mrjob.job import MRJob


class MRPassThruArgTest(MRJob):

    def configure_args(self):
        super(MRPassThruArgTest, self).configure_args()
        self.add_passthru_arg('--chars', action='store_true')
        self.add_passthru_arg('--ignore')

    def mapper(self, _, value):
        if self.options.ignore:
            value = value.replace(self.options.ignore, '')
        if self.options.chars:
            for c in value:
                yield c, 1
        else:
            for w in value.split(' '):
                yield w, 1

    def reducer(self, key, values):
        yield key, sum(values)


if __name__ == '__main__':
    MRPassThruArgTest.run()



# Copyright 2012 Yelp and Contributors
# Copyright 2013 Lyft
# Copyright 2014 Brett Gibson
# Copyright 2015-2017 Yelp
# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities related to cluster pooling. This code used to be in mrjob.emr.

In theory, this module might support pooling in general, but so far, there's
only a need for pooling on EMR.

"""
from collections import defaultdict
from logging import getLogger

from mrjob.aws import EC2_INSTANCE_TYPE_TO_COMPUTE_UNITS
from mrjob.aws import EC2_INSTANCE_TYPE_TO_MEMORY
from mrjob.py2 import integer_types

log = getLogger(__name__)

# we check the type and contents of requested fleets/groups because they
# are user-specified and may not have the correct format. Currently, we
# simply return no match, since either boto3 or the EMR AMI will catch
# the error when EMRJobRunner attempts to create a new cluster. See #1696


### identifying pooled clusters ###

def _pool_tags(hash, name):
    """Return a dict with "hidden" tags to add to the given cluster."""
    return dict(__mrjob_pool_hash=hash, __mrjob_pool_name=name)


def _extract_tags(cluster):
    """Pull the tags from a cluster, as a dict."""
    return {t['Key']: t['Value'] for t in cluster.get('Tags') or []}


def _pool_hash_and_name(cluster):
    """Return the hash and pool name for the given cluster, or
    ``(None, None)`` if it isn't pooled."""
    tags = _extract_tags(cluster)
    return tags.get('__mrjob_pool_hash'), tags.get('__mrjob_pool_name')


def _legacy_pool_hash_and_name(bootstrap_actions):
    """Get pool hash and name from a pre-v0.6.0 job."""
    for ba in bootstrap_actions:
        if ba['Name'] == 'master':
            args = ba['Args']
            if len(args) == 2 and args[0].startswith('pool-'):
                return args[0][5:], args[1]

    return None, None


### instance groups ###

def _instance_groups_satisfy(actual_igs, requested_igs):
    """Do the actual instance groups from a cluster satisfy the requested
    ones, for the purpose of pooling?
    """
    # the format of *requested_igs* is here:
    #     http://docs.aws.amazon.com/ElasticMapReduce/latest/API/API_InstanceGroup.html  # noqa
    # and the format of *actual_igs* is here:
    #     http://docs.aws.amazon.com/ElasticMapReduce/latest/API/API_ListInstanceGroups.html  # noqa

    # verify format of requested_igs
    if not (isinstance(requested_igs, (list, tuple)) and
            all(isinstance(req_ig, dict) and 'InstanceRole' in req_ig
                for req_ig in requested_igs)):
        log.debug('    bad instance_groups config')
        return None

    # a is a map from role to actual instance groups
    a = defaultdict(list)
    for ig in actual_igs:
        a[ig['InstanceGroupType']].append(ig)

    # r is a map from role to request (should be only one per role)
    r = {req.get('InstanceRole'): req for req in requested_igs}

    # updated request to account for extra instance groups
    # see #1630 for what we do when roles don't match
    if set(a) - set(r):
        r = _add_missing_roles_to_request(set(a) - set(r), r,
                                          ['InstanceCount'])

    if set(a) != set(r):
        log.debug("    missing instance group roles")
        return None

    sort_keys = {}
    for role in sorted(r):
        sort_key = _igs_for_same_role_satisfy(a[role], r[role])
        if not sort_key:  # doesn't satisfy
            return None
        sort_keys[role] = sort_key

    return tuple(sort_keys.get(role) for role in ('CORE', 'TASK', 'MASTER'))


def _igs_for_same_role_satisfy(actual_igs, requested_ig):
    """Does the *actual* list of instance groups satisfy the *requested*
    one?
    """
    # bid price/on-demand
    if not all(_ig_satisfies_bid_price(ig, requested_ig) for ig in actual_igs):
        return None

    # memory
    if not all(_ig_satisfies_mem(ig, requested_ig) for ig in actual_igs):
        return None

    # EBS volumes
    if not all(_ebs_satisfies(ig, requested_ig) for ig in actual_igs):
        return None

    # CPU (this returns # of compute units or None)
    return _igs_satisfy_cpu(actual_igs, requested_ig)


def _ig_satisfies_bid_price(actual_ig, requested_ig):
    """Does the actual instance group definition satisfy the bid price
    (or lack thereof) of the requested instance group?
    """
    # _instance_groups_satisfy() already verified *requested_ig* is a dict

    # on-demand instances satisfy every bid price
    if actual_ig['Market'] == 'ON_DEMAND':
        return True

    if requested_ig.get('Market', 'ON_DEMAND') == 'ON_DEMAND':
        log.debug('    spot instance, requested on-demand')
        return False

    if actual_ig['BidPrice'] == requested_ig.get('BidPrice'):
        return True

    try:
        if float(actual_ig['BidPrice']) >= float(requested_ig.get('BidPrice')):
            return True
        else:
            # low bid prices mean cluster is more likely to be
            # yanked away
            log.debug('    bid price too low')
            return False
    except ValueError:
        log.debug('    non-float bid price')
        return False


def _ig_satisfies_mem(actual_ig, requested_ig):
    """Does the actual instance group satisfy the memory requirements of
    the requested instance group?"""
    actual_type = actual_ig['InstanceType']
    requested_type = requested_ig.get('InstanceType')

    # this works even for unknown instance types
    if actual_type == requested_type:
        return True

    try:
        if (EC2_INSTANCE_TYPE_TO_MEMORY[actual_type] >=
                EC2_INSTANCE_TYPE_TO_MEMORY[requested_type]):
            return True
        else:
            log.debug('    too little memory')
            return False
    except KeyError:
        log.debug('    unknown instance type')
        return False


def _igs_satisfy_cpu(actual_igs, requested_ig):
    """Does the list of actual instance groups satisfy the CPU requirements
    of the requested instance group?

    If so, return the number of compute units. Otherwise, return ``None``

    If the requested instance type is unknown, just return the number
    of actual instances of the same type.
    """
    requested_type = requested_ig.get('InstanceType')
    num_requested = requested_ig.get('InstanceCount')

    if not isinstance(num_requested, integer_types):
        return None

    # count number of compute units (cu)
    if requested_type in EC2_INSTANCE_TYPE_TO_COMPUTE_UNITS:
        requested_cu = (
            num_requested * EC2_INSTANCE_TYPE_TO_COMPUTE_UNITS[requested_type])

        # don't require instances to be running; we'd be worse off if
        # we started our own cluster from scratch. (This can happen if
        # the previous job finished while some task instances were
        # still being provisioned.)
        actual_cu = sum(
            ig['RunningInstanceCount'] *
            EC2_INSTANCE_TYPE_TO_COMPUTE_UNITS.get(ig['InstanceType'], 0.0)
            for ig in actual_igs)
    else:
        # unknown instance type, just count # of matching instances
        requested_cu = num_requested
        actual_cu = sum(ig['RunningInstanceCount'] for ig in actual_igs
                        if ig['InstanceType'] == requested_type)

    if actual_cu >= requested_cu:
        return actual_cu
    else:
        log.debug('    not enough compute units')
        return None


### instance fleets ###

def _instance_fleets_satisfy(actual_fleets, req_fleets):
    """Common code for :py:func:`
    :py:func:`_instance_groups_satisfy_fleets` and
    :py:func:`_instance_groups_satisfy`."""
    # verify format of requested_igs
    if not (isinstance(req_fleets, (list, tuple)) and
            all(isinstance(req_ft, dict) and 'InstanceFleetType' in req_ft
                for req_ft in req_fleets)):
        log.debug('    bad instance_fleets config')
        return None

    # a is a map from role to actual instance fleet
    # (unlike with groups, there can never be more than one fleet per role)
    a = {f['InstanceFleetType']: f for f in actual_fleets}

    # r is a map from role to request (should be only one per role)
    r = {f['InstanceFleetType']: f for f in req_fleets}

    # updated request to account for extra instance groups
    # see #1630 for what we do when roles don't match
    if set(a) - set(r):
        r = _add_missing_roles_to_request(
            set(a) - set(r), r,
            ['TargetOnDemandCapacity', 'TargetSpotCapacity'])

    if set(a) != set(r):
        log.debug("    missing instance fleet roles")
        return None

    sort_keys = {}
    for role in sorted(r):
        sort_key = _fleet_for_same_role_satisfies(a[role], r[role])
        if not sort_key:  # doesn't satisfy
            return None
        sort_keys[role] = sort_key

    return tuple(sort_keys.get(role) for role in ('CORE', 'TASK', 'MASTER'))


def _fleet_for_same_role_satisfies(actual_fleet, req_fleet):
    # match up instance types
    actual_specs = {spec['InstanceType']: spec
                    for spec in actual_fleet['InstanceTypeSpecifications']}
    try:
        req_specs = {spec['InstanceType']: spec
                     for spec in req_fleet['InstanceTypeConfigs']}
    except (TypeError, KeyError):
        return

    if set(actual_specs) - set(req_specs):
        log.debug('    fleet may include wrong instance types')
        return

    if not all(_fleet_spec_satsifies(actual_specs[t], req_specs[t])
               for t in actual_specs):
        return

    # capacity
    actual_on_demand = actual_fleet.get('ProvisionedOnDemandCapacity', 0)
    req_on_demand = req_fleet.get('TargetOnDemandCapacity', 0)

    if not isinstance(req_on_demand, integer_types):
        return

    if req_on_demand > actual_on_demand:
        log.debug('    not enough on-demand capacity')
        return

    actual_spot = actual_fleet.get('ProvisionedSpotCapacity', 0)
    req_spot = req_fleet.get('TargetSpotCapacity', 0)

    if not isinstance(req_spot, integer_types):
        return

    # allow extra on-demand instances to serve as spot instances
    if req_spot > actual_spot + (actual_on_demand - req_on_demand):
        log.debug('    not enough spot capacity')

    # handle TERMINATE_CLUSTER timeout action. This really doesn't play
    # well with pooling anyhow
    if _get_timeout_action(actual_fleet) == 'TERMINATE_CLUSTER':
        if _get_timeout_action(req_fleet) != 'TERMINATE_CLUSTER':
            log.debug('    self-terminating fleet not requested')
            return

        if (_get_timeout_duration(actual_fleet) <
                _get_timeout_duration(req_fleet)):
            log.debug('    fleet may self-terminate prematurely')
            return

    # sort by total capacity, with on-demand as a tie-breaker
    return (actual_on_demand + actual_spot, actual_on_demand)


def _get_timeout_action(fleet):
    return fleet.get(
        'LaunchSpecifications', {}).get(
        'SpotSpecification', {}).get(
        'TimeoutAction')


def _get_timeout_duration(fleet):
    return fleet.get(
        'LaunchSpecifications', {}).get(
        'SpotSpecification', {}).get(
        'TimeoutDurationMinutes', 0.0)


def _fleet_spec_satsifies(actual_spec, req_spec):
    """Make sure the specification for the given instance type is as
    good or better than the requested spec.

    Specs must have the same weight, but "better" EBS configurations are
    accepted.

    Bid price must either be higher or the *actual* bid price
    must be same as on-demand
    """
    if (actual_spec.get('WeightedCapacity', 1) !=
            req_spec.get('WeightedCapacity', 1)):
        log.debug('    different weighted capacity for same instance type')
        return

    if not _ebs_satisfies(actual_spec, req_spec):
        return

    # bid price is the max, don't worry about it
    if actual_spec.get('BidPriceAsPercentageOfOnDemandPrice', 100) >= 100:
        return True

    # absolute bid price
    req_bid_price = req_spec.get('BidPrice')
    if req_bid_price is not None:
        actual_bid_price = actual_spec.get('BidPrice')
        if actual_bid_price is None:
            log.debug('    no bid price specified')
            return

        try:
            if not float(actual_bid_price) >= float(req_bid_price):
                log.debug('    bid price too low')
                return
        except TypeError:
            log.debug('    non-numeric bid price')
            return

    # relative bid price
    req_bid_percent = req_spec.get('BidPriceAsPercentageOfOnDemandPrice')
    if not isinstance(req_spec, (integer_types, float)):
        return

    if req_bid_percent:
        actual_bid_percent = actual_spec.get(
            'BidPriceAsPercentageOfOnDemandPrice')
        if actual_bid_percent is None:
            log.debug('    no bid price as % of on-demand price')
            return

        if req_bid_percent > actual_bid_percent:
            log.debug('    bid price as % of on-demand price too low')
            return

    return True


### common code for matching instance (groups or fleets) ###


def _add_missing_roles_to_request(
        missing_roles, role_to_req, req_count_fields):
    """Helper for :py:func:`_igs_satisfy_request`. Add requests for
    *missing_roles* to *role_to_ig* so that we have a better chance of
    matching the cluster's actual instance groups."""
    # see #1630 for discussion

    # don't worry about modifying *role_to_req*; this is
    # a helper func

    if 'CORE' in missing_roles and list(role_to_req) == ['MASTER']:
        # both core and master have to satisfy master-only request
        role_to_req['CORE'] = role_to_req['MASTER']

    if 'TASK' in missing_roles and 'CORE' in role_to_req:
        # make sure tasks won't crash on the task instances,
        # but don't require the same amount of CPU
        role_to_req['TASK'] = dict(role_to_req['CORE'])
        for req_count_field in req_count_fields:
            role_to_req['TASK'][req_count_field] = 0

    return role_to_req


def _ebs_satisfies(actual, request):
    """Does *actual* have EBS volumes that satisfy *request*.

    *actual* is either an instance group from ``ListInstanceGroups``
    or an instance fleet spec from ``ListInstanceFleets`` (format
    is the same).

    *request* is either the ``InstanceGroups`` or ``InstanceFleets``
    param to ``RunJobFlow``

    If *request* doesn't have an EBS Configuration, we return
    True.

    If *request* requests EBS optimization, *actual* should provide it.

    Finally, *actual* should have the same or better block devices
    as those in *request* (same volume type, at least as much IOPS
    and volume size).
    """
    req_ebs_config = request.get('EbsConfiguration')

    if not req_ebs_config:
        return True

    if (req_ebs_config.get('EbsOptimized') and
            not actual.get('EbsOptimized')):
        log.debug('    need EBS-optimized instances')
        return False

    req_device_configs = req_ebs_config.get('EbsBlockDeviceConfigs')

    if not req_device_configs:
        return True

    if not (isinstance(req_device_configs, (list, tuple)) and
            all(isinstance(rdc, dict) for rdc in req_device_configs)):
        return False

    req_volumes = []

    for req_device_config in req_device_configs:
        volume = req_device_config['VolumeSpecification']
        num_volumes = req_device_config.get('VolumesPerInstance', 1)

        req_volumes.extend([volume] * num_volumes)

    actual_volumes = [
        bd.get('VolumeSpecification', {})
        for bd in actual.get('EbsBlockDevices', [])]

    return _ebs_volumes_satisfy(actual_volumes, req_volumes)


def _ebs_volumes_satisfy(actual_volumes, req_volumes):
    """Does the given list of actual EBS volumes satisfy the given request?

    Just compare them one by one (we want each actual device to be
    bigger/faster; just having the same amount of capacity or iops
    isn't enough).
    """
    if not isinstance(req_volumes, (list, tuple)):
        return False

    if len(req_volumes) > len(actual_volumes):
        log.debug('    more EBS volumes requested than available')
        return False

    return all(_ebs_volume_satisfies(a, r)
               for a, r in zip(actual_volumes, req_volumes))


def _ebs_volume_satisfies(actual_volume, req_volume):
    """Does the given actual EBS volume satisfy the given request?"""
    if not isinstance(req_volume, dict):
        return False

    if req_volume.get('VolumeType') != actual_volume.get('VolumeType'):
        log.debug('    wrong EBS volume type')
        return False

    if not req_volume.get('SizeInGB', 0) <= actual_volume.get('SizeInGB', 0):
        log.debug('    EBS volume too small')
        return False

    # Iops isn't really "optional"; it has to be set if volume type is
    # io1 and not set otherwise
    if not (req_volume.get('Iops', 0) <= actual_volume.get('Iops', 0)):
        log.debug('    EBS volume too slow')
        return False

    return True



# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from os.path import join
import os
from os.path import join

from mrjob.job import MRJob


class MRUploadAttrsJob(MRJob):
    """Use the FILES, DIRS, and ARCHIVES attrs, and list contents of
    the current directory"""

    FILES = ['mr_upload_attrs_job/README.txt',
             'mr_upload_attrs_job/empty.tar.gz']

    ARCHIVES = ['mr_upload_attrs_job/empty.tar.gz#empty']

    DIRS = ['mr_upload_attrs_job']

    def mapper(self, key, value):
        pass

    def mapper_final(self):
        for dirpath, _, filenames in os.walk('.', followlinks=True):
            yield dirpath, None
            for filename in filenames:
                yield join(dirpath, filename), None

    def reducer(self, key, values):
        # remove duplicates
        for value in values:
            yield (key, value)
            break


if __name__ == '__main__':
    MRUploadAttrsJob.run()



# Copyright 2009-2012 Yelp
# Copyright 2013 Tom Arnfeld and David Marin
# Copyright 2014 Contributors
# Copyright 2015-2016 Yelp
# Copyright 2017 Yelp and Contributors
# Copyright 2018 Yelp
# Copyright 2019 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A mock version of the hadoop binary that actually manipulates the
filesystem. This imitates only things that mrjob actually uses.

Requires these environment variables:
  MOCK_HADOOP_TMP: temp directory used by mock hadoop
  MOCK_HADOOP_VERSION: version of Hadoop to emulate (e.g. '2.7.1').

This is designed to run as: python -m tests.mockhadoop <hadoop args>

mrjob requires a single binary (no args) to stand in for hadoop, so
use create_mock_hadoop_script() to write out a shell script that runs
mockhadoop.

To add mock output for streaming steps, use add_mock_output()
To add mock counters for steps, use add_mock_counters()
To see how the hadoop binary has been invoked, use get_mock_hadoop_cmd_args()
To see what's in HDFS, use get_mock_hdfs_root()
"""
from __future__ import print_function

import datetime
import glob
import json
import os
import os.path
import pipes
import shutil
import stat
import sys

from mrjob.compat import uses_yarn
from mrjob.hadoop import _HADOOP_STREAMING_JAR_RE
from mrjob.parse import urlparse
from mrjob.util import shlex_split

from tests.sandbox import SandboxedTestCase


class MockHadoopTestCase(SandboxedTestCase):

    def setUp(self):
        super(MockHadoopTestCase, self).setUp()
        # setup fake hadoop home
        hadoop_home = self.makedirs('mock_hadoop_home')
        os.environ['HADOOP_HOME'] = hadoop_home
        os.environ['MOCK_HADOOP_VERSION'] = "1.2.0"
        os.environ['MOCK_HADOOP_TMP'] = self.makedirs('mock_hadoop_tmp')

        # make fake hadoop binary
        os.mkdir(os.path.join(hadoop_home, 'bin'))
        self.hadoop_bin = os.path.join(hadoop_home, 'bin', 'hadoop')
        create_mock_hadoop_script(self.hadoop_bin)

        # make fake streaming jar
        os.makedirs(os.path.join(hadoop_home, 'contrib', 'streaming'))
        streaming_jar_path = os.path.join(
            hadoop_home, 'contrib', 'streaming', 'hadoop-0.X.Y-streaming.jar')
        open(streaming_jar_path, 'w').close()

        # make sure the fake hadoop binaries can find mrjob
        self.add_mrjob_to_pythonpath()


# layout of $MOCK_HADOOP_TMP:
#   cmd.log: single file containing one line per invocation of mock hadoop
#     binary, with lines containing space-separated args passed
#   counters/YYmmdd.HH:MM:SS.uuuuuu.json: files containing JSON map from
#     group to counter to amount. When we need to output counters, we read
#     the first file alphabetically, and then delete it
#   hdfs/: The root directory of all filesystems (HDFS and otherwise) accessed
#     by mock hadoop binary (we ignore scheme and host)
#   output/YYmmdd.HH:MM:SS.uuuuuu/part-nnnnn: sudirectories containing
#     part files containing output to copy into HDFS to simulate job output.
#     When we need output, we pick the first directory alphabetically, and then
#     delete it


def create_mock_hadoop_script(path):
    """Dump a wrapper script to the given file object that runs this
    python script."""
    # make this work even if $PATH or $PYTHONPATH changes
    with open(path, 'w') as f:
        f.write('#!/bin/sh\n')
        f.write('%s %s "$@"\n' % (
            pipes.quote(sys.executable),
            pipes.quote(os.path.abspath(__file__))))
    os.chmod(path, stat.S_IREAD | stat.S_IEXEC)


def iso_now():
    """Return the current time as YYmmdd.HH:MM:SS.uuuuuu. Use this to ensure
    that mock data added stays in alphabetical order.
    """
    now = datetime.datetime.now()
    return '%s.%06d' % (now.strftime('%Y%m%d.%H%M%S'), now.microsecond)


def get_mock_dir(*subdirs, **kwargs):
    """Get a directory within $MOCK_HADOOP_TMP, creating it if it doesn't
    already exist.

    Takes optional *environ* keyword arg (defaults to ``os.environ``)."""
    environ = kwargs.get('environ') or os.environ

    path = os.path.join(environ['MOCK_HADOOP_TMP'], *subdirs)

    if not os.path.exists(path):
        os.makedirs(os.path.abspath(path))

    return path


def add_mock_hadoop_output(parts):
    """Add mock output which will be used by the next fake streaming
    job that mockhadoop will run.

    Args:
    parts -- a list of the contents of parts files, which should be iterables
        that return lines as bytes (e.g. lists, BytesIOs).
    """
    output_subdir = get_mock_dir('output', iso_now())

    for i, part in enumerate(parts):
        part_path = os.path.join(output_subdir, 'part-%05d' % i)
        with open(part_path, 'wb') as part_file:
            part_file.write(part)


def get_mock_hadoop_output():
    """Get a path to a directory containing part files to use as mock output.
    """
    output_dir = get_mock_dir('output')

    dirnames = sorted(os.listdir(output_dir))
    if dirnames:
        return os.path.join(output_dir, dirnames[0])
    else:
        return None


def add_mock_hadoop_counters(counters):
    """Add counters for a subsequent run of the hadoop jar command (streaming
    or otherwise), as a map from group name to counter name to amount."""
    counter_path = os.path.join(get_mock_dir('counters'), iso_now() + '.json')

    with open(counter_path, 'w') as f:
        json.dump(counters, f)


def next_mock_hadoop_counters():
    """Read the next set of counters from the filesystem, and delete it.

    If no counters available, return {}
    """
    counters_dir = get_mock_dir('counters')
    filenames = sorted(os.listdir(counters_dir))

    if filenames:
        counters_path = os.path.join(counters_dir, filenames[0])
        try:
            with open(counters_path) as f:
                return json.load(f)
        finally:
            os.remove(counters_path)
    else:
        return {}


def get_mock_hdfs_root(environ=None):
    """Get the path of mock root of HDFS. Creates the directory if it
    doesn't already exist."""
    return get_mock_dir('hdfs', environ=environ)


def mock_hadoop_uses_yarn(environ):
    return uses_yarn(environ['MOCK_HADOOP_VERSION'])


def get_mock_hadoop_cmd_args():
    """Get a list for each invocation of hadoop, each containing a list of
    arguments (not including the hadoop binary's path)."""
    cmd_log = os.path.join(get_mock_dir(), 'cmd.log')

    if not os.path.exists(cmd_log):
        return []

    with open(cmd_log) as f:
        return [shlex_split(cmd) for cmd in f]


def hdfs_uri_to_real_path(hdfs_uri, environ):
    """Map an HDFS URI to a path on the filesystem."""
    components = urlparse(hdfs_uri)

    scheme = components.scheme
    path = components.path

    if not scheme and not path.startswith('/'):
        path = '/user/%s/%s' % (environ['USER'], path)

    return os.path.join(get_mock_hdfs_root(environ=environ), path.lstrip('/'))


def real_path_to_hdfs_uri(real_path, environ):
    """Map a real path to an hdfs:/// URI."""
    hdfs_root = get_mock_hdfs_root(environ=environ)

    if not real_path.startswith(hdfs_root):
        raise ValueError('path %s is not in %s' % (real_path, hdfs_root))

    # for some reason, relpath() doesn't work here
    hdfs_uri = real_path[len(hdfs_root):]
    if not hdfs_uri.startswith('/'):
        hdfs_uri = '/' + hdfs_uri

    return hdfs_uri


def invoke_cmd(stdout, stderr, environ, prefix, cmd, cmd_args, error_msg,
               error_status):
    """Helper function to call command and subcommands of the hadoop binary.

    Basically, combines prefix and cmd to make a function name, and calls
    it with cmd_args. If no such function exists, prints error_msg
    to stderr, and exits with status error_status.
    """
    func_name = prefix + cmd

    if func_name in globals():
        return globals()[func_name](stdout, stderr, environ, *cmd_args)
    else:
        stderr.write(error_msg)
        return -1


def main(stdin, stdout, stderr, argv, environ):
    """Implements hadoop <args>"""
    # log what commands we ran
    cmd_log_path = os.path.join(get_mock_dir(environ=environ), 'cmd.log')
    with open(cmd_log_path, 'a') as cmd_log:
        cmd_log.write(' '.join(pipes.quote(arg) for arg in argv[1:]))
        cmd_log.write('\n')
        cmd_log.flush()

    if len(argv) < 2:
        print('Usage: hadoop [--config confdir] COMMAND\n', file=stderr)
        return 1

    cmd = argv[1]
    cmd_args = argv[2:]

    return invoke_cmd(
        stdout, stderr, environ, 'hadoop_', cmd, cmd_args,
        'Could not find the main class: %s.  Program will exit.\n\n' % cmd, 1)


def hadoop_fs(stdout, stderr, environ, *args):
    """Implements hadoop fs <args>"""
    if len(args) < 1:
        print('Usage: java FsShell', file=stderr)
        return -1

    cmd = args[0][1:]  # convert e.g. '-put' -> 'put'
    cmd_args = args[1:]

    # this doesn't have to be a giant switch statement, but it's a
    # bit easier to understand this way. :)
    return invoke_cmd(stdout, stderr, environ, 'hadoop_fs_', cmd, cmd_args,
                      '%s: Unknown command\nUsage: java FsShell\n' % cmd, -1)


def hadoop_fs_cat(stdout, stderr, environ, *args):
    """Implements hadoop fs -cat <src>"""
    if len(args) < 1:
        print('Usage: java FsShell [-cat <src>]', file=stderr)
        return -1

    failed = False
    for hdfs_uri_glob in args:
        real_path_glob = hdfs_uri_to_real_path(hdfs_uri_glob, environ)
        paths = glob.glob(real_path_glob)
        if not paths:
            print('cat: File does not exist: %s' % hdfs_uri_glob, file=stderr)
            failed = True
        else:
            for path in paths:
                with open(path, 'rb') as f:
                    # use binary interface if available
                    stdout_buffer = getattr(stdout, 'buffer', stdout)
                    for line in f:
                        stdout_buffer.write(line)

    if failed:
        return -1
    else:
        return 0


def hadoop_fs_du(stdout, stderr, environ, *args):
    if mock_hadoop_uses_yarn(environ) and args and args[0] == '-s':
        aggregate = True
        path_args = args[1:]
    else:
        aggregate = False
        path_args = args

    return _hadoop_fs_du('du', stdout, stderr, environ,
                         path_args=path_args, aggregate=aggregate)


def hadoop_fs_dus(stdout, stderr, environ, *args):
    """Implements hadoop fs -dus."""
    yarn = mock_hadoop_uses_yarn(environ)

    if yarn:
        print("dus: DEPRECATED: Please use 'du -s' instead.", file=stderr)

    return _hadoop_fs_du('dus', stdout, stderr, environ, path_args=args,
                         aggregate=True, pre_yarn_dus_format=not yarn)


def _hadoop_fs_du(cmd_name, stdout, stderr, environ, path_args,
                  aggregate, pre_yarn_dus_format=False):
    """Implements fs -du and fs -dus."""
    hdfs_uri_globs = path_args or ['']

    failed = False

    for hdfs_uri_glob in hdfs_uri_globs:
        real_path_glob = hdfs_uri_to_real_path(hdfs_uri_glob, environ)
        real_paths = glob.glob(real_path_glob)
        if not real_paths:
            if mock_hadoop_uses_yarn(environ):
                msg = "%s: `%s': No such file or directory" % (
                    cmd_name, hdfs_uri_glob)
            else:
                msg = '%s: Cannot access %s: No such file or directory.' % (
                    cmd_name, hdfs_uri_glob)
            print(msg, file=stderr)

            failed = True
            continue

        for real_path in real_paths:
            # if we're not in -s mode, it finds size of every item in dir
            if os.path.isdir(real_path) and not aggregate:
                paths = [os.path.join(real_path, name)
                         for name in os.listdir(real_path)]
            else:
                paths = [real_path]

            for path in paths:
                hdfs_uri = real_path_to_hdfs_uri(path, environ)
                size = _du(path)

                # prior to YARN, the du commands put a variable number
                # of spaces between path and size. Not emulating this
                # because it doesn't matter.

                if pre_yarn_dus_format:  # old fs -dus only, not fs -du
                    print('%s  %d' % (hdfs_uri, size), file=stdout)
                else:
                    print('%d  %s' % (size, hdfs_uri), file=stdout)

    if failed:
        if mock_hadoop_uses_yarn(environ):
            return 1  # -1 is only for arg-parsing errors
        else:
            return -1
    else:
        return 0


def _du(real_path):
    """Get total size of file or files in dir (recursive)."""
    # Hadoop also seems to have some rule for size of directory
    # that varies across versions and that I don't fully understand.
    # Ignoring this for testing.
    total_size = 0

    if os.path.isdir(real_path):
        for dirpath, dirnames, filenames in os.walk(real_path):
            for filename in filenames:
                total_size += os.path.getsize(
                    os.path.join(dirpath, filename))
    else:
        total_size += os.path.getsize(real_path)

    return total_size


def _hadoop_ls_line(real_path, scheme, netloc, size=0, max_size=0, environ={}):
    hdfs_uri = real_path_to_hdfs_uri(real_path, environ)

    # we could actually implement ls here, but mrjob only cares about
    # the path
    if os.path.isdir(real_path):
        file_type = 'd'
    else:
        file_type = '-'

    if scheme in ('s3', 's3n', 's3a'):
        # no user and group on S3 (see Pull Request #573)
        user_and_group = ''
    else:
        user_and_group = 'dave supergroup'

    # newer Hadoop returns fully qualified URIs (see Pull Request #577)
    if scheme and mock_hadoop_uses_yarn(environ):
        hdfs_uri = '%s://%s%s' % (scheme, netloc, hdfs_uri)

    # figure out the padding
    size = str(size).rjust(len(str(max_size)))

    return (
        '%srwxrwxrwx - %s %s 2010-10-01 15:16 %s' %
        (file_type, user_and_group, size, hdfs_uri))


def hadoop_fs_ls(stdout, stderr, environ, *args):
    """Implements hadoop fs -ls."""
    if mock_hadoop_uses_yarn(environ) and args and args[0] == '-R':
        path_args = args[1:]
        recursive = True
    else:
        path_args = args
        recursive = False

    return _hadoop_fs_ls('ls', stdout, stderr, environ,
                         path_args=path_args, recursive=recursive)


def hadoop_fs_lsr(stdout, stderr, environ, *args):
    """Implements hadoop fs -lsr."""
    if mock_hadoop_uses_yarn(environ):
        print("lsr: DEPRECATED: Please use 'ls -R' instead.", file=stderr)

    return _hadoop_fs_ls(
        'lsr', stdout, stderr, environ, path_args=args, recursive=True)


def _hadoop_fs_ls(cmd_name, stdout, stderr, environ, path_args, recursive):
    """Helper for hadoop_fs_ls() and hadoop_fs_lsr()."""
    hdfs_uri_globs = path_args or ['']

    failed = False
    for hdfs_uri_glob in hdfs_uri_globs:
        parsed = urlparse(hdfs_uri_glob)
        scheme = parsed.scheme
        netloc = parsed.netloc

        real_path_glob = hdfs_uri_to_real_path(hdfs_uri_glob, environ)
        real_paths = glob.glob(real_path_glob)

        paths = []

        if not real_paths:
            print('%s: Cannot access %s: No such file or directory.' %
                  (cmd_name, hdfs_uri_glob), file=stderr)
            failed = True
        else:
            for real_path in real_paths:
                if os.path.isdir(real_path):
                    if recursive:
                        for dirpath, dirnames, filenames in os.walk(real_path):
                            paths.append((dirpath, scheme, netloc, 0))
                            for filename in filenames:
                                path = os.path.join(dirpath, filename)
                                size = os.path.getsize(path)
                                paths.append((path, scheme, netloc, size))
                    else:
                        for filename in os.listdir(real_path):
                            path = os.path.join(real_path, filename)
                            if os.path.isdir(path):
                                size = 0
                            else:
                                size = os.path.getsize(path)
                            paths.append((path, scheme, netloc, size))
                else:
                    size = os.path.getsize(real_path)
                    paths.append((real_path, scheme, netloc, size))

        if paths:
            print('Found %d items' % len(paths), file=stdout)
            max_size = max(size for _, __, ___, size in paths)
            for path in paths:
                print(_hadoop_ls_line(*path + (max_size, environ)),
                      file=stdout)

    if failed:
        return -1
    else:
        return 0


def hadoop_fs_mkdir(stdout, stderr, environ, *args):
    """Implements hadoop fs -mkdir"""
    if len(args) < 1:
        print('Usage: java FsShell [-mkdir <path>]', file=stderr)
        return -1

    failed = False

    if mock_hadoop_uses_yarn(environ):
        # expect a -p parameter for mkdir
        if args[0] == '-p':
            args = args[1:]
        else:
            failed = True

    for path in args:
        real_path = hdfs_uri_to_real_path(path, environ)
        if os.path.exists(real_path):
            print('mkdir: cannot create directory %s: File exists' % path,
                  file=stderr)
            # continue to make directories on failure
            failed = True
        else:
            os.makedirs(real_path)

    if failed:
        return -1
    else:
        return 0


def hadoop_fs_put(stdout, stderr, environ, *args):
    """Implements hadoop fs -put"""
    if len(args) < 2:
        print('Usage: java FsShell [-put <localsrc> ... <dst>]', file=stderr)
        return -1

    srcs = args[:-1]
    dst = args[-1]

    real_dst = hdfs_uri_to_real_path(dst, environ)
    real_dir = os.path.dirname(real_dst)
    # dst could be a dir or a filename; we don't know
    if not (os.path.isdir(real_dst) or os.path.isdir(real_dir)):
        os.makedirs(real_dir)

    for src in srcs:
        shutil.copy(src, real_dst)
    return 0


def hadoop_fs_rm(stdout, stderr, environ, *args):
    """Implements hadoop fs -rm."""
    # parse args
    recursive = False
    force = False

    yarn = mock_hadoop_uses_yarn(environ)

    for i, arg in enumerate(args):
        # -r/-R and -f are only available in YARN
        if yarn and arg in ('-r', '-R'):
            recursive = True
        elif yarn and arg == '-f':
            force = True
        elif arg == '-skipTrash':
            pass  # we don't emulate trash
        elif arg.startswith('-'):
            # don't know what the pre-YARN version of this, doesn't matter
            # It's really '-rm', not 'rm'. Because it's about args?
            print('-rm: Illegal option %s' % arg, file=stderr)
            return -1
        else:
            # BSD-style args: all switches at the beginning
            path_args = args[i:]
            break
    else:
        # no path arguments
        if yarn:
            print('-rm: Not enough arguments: expected 1 but got 0',
                  file=stderr)
            print('Usage: hadoop fs [generic options] -rm [-f] [-r|-R]'
                  ' [-skipTrash] <src> ...', file=stderr)
        else:
            print('Usage: java FsShell [-rm [-skipTrash] <src>]', file=stderr)

        return -1

    return _hadoop_fs_rm('rm', stdout, stderr, environ,
                         path_args=path_args, recursive=recursive,
                         force=force)


def hadoop_fs_rmr(stdout, stderr, environ, *args):
    yarn = mock_hadoop_uses_yarn(environ)

    if yarn:
        print("rmr: DEPRECATED: Please use 'rm -r' instead.", file=stderr)

    if args and args[0] == '-skipTrash':
        path_args = args[1:]
    else:
        path_args = args

    if not path_args:
        if yarn:
            print('-rmr: Not enough arguments: expected 1 but got 0',
                  file=stderr)
            # -skipTrash isn't mentioned in usage, at least in 2.5.2
            print('Usage: hadoop fs [generic options] -rmr'
                  ' <src> ...', file=stderr)
        else:
            print('Usage: java FsShell [-rmr [-skipTrash] <src>]', file=stderr)

        return -1

    return _hadoop_fs_rm('rmr', stdout, stderr, environ,
                         path_args=path_args, recursive=True, force=False)


def _hadoop_fs_rm(cmd_name, stdout, stderr, environ,
                  path_args, recursive, force):
    """Helper for hadoop_fs_rm() and hadoop_fs_rmr()."""
    # use an array so that fail() can update it
    failed = []

    def fail(path, msg):
        if mock_hadoop_uses_yarn(environ):
            print('%s `%s`: %s' % (cmd_name, path, msg), file=stderr)
        else:
            print('%s: cannot remove %s: %s.' % (cmd_name, path, msg),
                  file=stderr)
        failed.append(True)

    for path in path_args:
        real_path = hdfs_uri_to_real_path(path, environ)
        if os.path.isdir(real_path):
            if recursive:
                shutil.rmtree(real_path)
            else:
                fail(path, 'Is a directory')
        elif os.path.exists(real_path):
            os.remove(real_path)
        else:
            if not force:
                fail(path, 'No such file or directory')

    if failed:
        return -1
    else:
        return 0


def hadoop_fs_test(stdout, stderr, environ, *args):
    """Implements hadoop fs -test."""
    if len(args) < 1:
        print('Usage: java FsShell [-test -[ezd] <src>]', file=stderr)

    if os.path.exists(hdfs_uri_to_real_path(args[1], environ)):
        return 0
    else:
        return 1


def hadoop_fs_touchz(stdout, stderr, environ, *args):
    """Implements hadoop fs -touchz'"""
    if len(args) < 1:
        print('Usage: hadoop fs [generic options] -touchz <path> ...]',
              file=stderr)
        return -1

    uri = args[0]
    path = hdfs_uri_to_real_path(uri, environ)

    if os.path.exists(path):
        if os.path.getsize(path) == 0:
            return 0
        else:
            print(
                "touchz: `%s': Not a zero-length file" % uri, file=stderr)
            return 1
    else:
        if os.path.exists(os.path.dirname(path)):
            with open(path, 'w'):
                pass
            return 0
        else:
            print(
                "touchz: `%s': No such file or directory: `%s'" % (
                    uri, uri), file=stderr)
            return 1


def hadoop_jar(stdout, stderr, environ, *args):
    if len(args) < 1:
        print('RunJar jarFile [mainClass] args...', file=stderr)
        return -1

    jar_path = args[0]
    if not os.path.exists(jar_path):
        print('Exception in thread "main" java.io.IOException: Error opening'
              ' job jar: %s' % jar_path, file=stderr)
        return -1

    # use this to simulate log4j
    def mock_log4j(message, level='INFO', logger='mapreduce.JOB', now=None):
        now = now or datetime.datetime.now()
        line = '%s %s %s: %s' % (now.strftime('%Y/%m/%d %H:%M:%S'),
                                 level, logger, message)
        print(line, file=stderr)

    # simulate counters
    counters = next_mock_hadoop_counters()
    if counters:
        num_counters = sum(len(g) for g in counters.values())
        mock_log4j('Counters: %d' % num_counters)
        # subsequent lines are actually part of same log record
        for group, group_counters in sorted(counters.items()):
            print(('\t%s' % group), file=stderr)
            for counter, amount in sorted(group_counters.items()):
                print(('\t\t%s=%d' % (counter, amount)), file=stderr)

    # simulate output for streaming steps
    if _HADOOP_STREAMING_JAR_RE.match(os.path.basename(jar_path)):
        streaming_args = args[1:]
        output_idx = list(streaming_args).index('-output')
        assert output_idx != -1
        output_dir = streaming_args[output_idx + 1]
        real_output_dir = hdfs_uri_to_real_path(output_dir, environ)

        mock_output_dir = get_mock_hadoop_output()
        if mock_output_dir is None:
            mock_log4j('Job failed!')
            return -1

        if os.path.isdir(real_output_dir):
            os.rmdir(real_output_dir)

        shutil.move(mock_output_dir, real_output_dir)

    now = datetime.datetime.now()
    mock_log4j(now.strftime('Running job: job_%Y%m%d%H%M_0001'))
    mock_log4j('Job succeeded!')
    return 0


def hadoop_version(stdout, stderr, environ, *args):
    stdout.write("Hadoop " + environ['MOCK_HADOOP_VERSION'])
    return 0


if __name__ == '__main__':
    sys.exit(main(sys.stdin, sys.stdout, sys.stderr, sys.argv, os.environ))



# Copyright 2009-2012 Yelp and Contributors
# Copyright 2013 David Marin
# Copyright 2014-2016 Yelp and Contributors
# Copyright 2017-2018 Yelp
# Copyright 2019 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import codecs
import logging
import os
import sys
import time
from io import BytesIO
from argparse import ArgumentParser
from argparse import ArgumentTypeError

from mrjob.conf import combine_dicts
from mrjob.conf import combine_lists
from mrjob.local import LocalMRJobRunner
from mrjob.options import _RUNNER_OPTS
from mrjob.options import _add_basic_args
from mrjob.options import _add_job_args
from mrjob.options import _add_runner_args
from mrjob.options import _optparse_kwargs_to_argparse
from mrjob.options import _parse_raw_args
from mrjob.options import _print_help_for_runner
from mrjob.options import _print_basic_help
from mrjob.runner import _runner_class
from mrjob.setup import parse_legacy_hash_path
from mrjob.step import StepFailedException
from mrjob.util import log_to_null
from mrjob.util import log_to_stream


log = logging.getLogger(__name__)


# sentinel value; used when running MRJob as a script
_READ_ARGS_FROM_SYS_ARGV = '_READ_ARGS_FROM_SYS_ARGV'


def _im_func(f):
    """Wrapper to get at the underlying function belonging to a method.

    Python 2 is slightly different because classes have "unbound methods"
    which wrap the underlying function, whereas on Python 3 they're just
    functions. (Methods work the same way on both versions.)
    """
    # "im_func" is the old Python 2 name for __func__
    if hasattr(f, '__func__'):
        return f.__func__
    else:
        return f


class MRJobLauncher(object):
    """Handle running a MapReduce job on an executable from the command line.
    This class will eventually support running arbitrary executables; for now
    it only supports :py:class:`~mrjob.job.MRJob` subclasses. Up to v0.5 it is
    effectively part of the :py:class:`~mrjob.job.MRJob` class itself and
    should not be used externally in any way.
    """
    # only MRJobLauncher expects the first argument to be script_path
    _FIRST_ARG_IS_SCRIPT_PATH = True

    def __init__(self, script_path=None, args=None, from_cl=False):
        """
        :param script_path: Path to script unless it's the first item of *args*
        :param args: Command line arguments
        :param from_cl: If not using sys.argv but still comming from the
                        command line (as opposed to a script, e.g. from
                        mrjob.cmd), don't override the option parser error
                        function (exit instead of throwing ValueError).
        """
        if script_path is not None:
            script_path = os.path.abspath(script_path)
        self._script_path = script_path

        # make sure we respect the $TZ (time zone) environment variable
        if hasattr(time, 'tzset'):
            time.tzset()

        # argument dests for args to pass through
        self._passthru_arg_dests = set()
        self._file_arg_dests = set()

        # there is no equivalent in argparse
        # remove this in v0.7.0
        if hasattr(self, 'OPTION_CLASS'):
            log.warning('OPTION_CLASS attribute is ignored; '
                        'mrjob now uses argparse instead of optparse')

        self.arg_parser = ArgumentParser(usage=self._usage(),
                                         add_help=False)
        self.configure_args()

        if (_im_func(self.configure_options) !=
                _im_func(MRJobLauncher.configure_options)):
            log.warning('configure_options() is deprecated and will be'
                        ' removed in v0.7.0; please use configure_args()'
                        ' instead.')
            self.configure_options()

        # don't pass None to parse_args unless we're actually running
        # the MRJob script
        if args is _READ_ARGS_FROM_SYS_ARGV:
            self._cl_args = sys.argv[1:]
        else:
            # don't pass sys.argv to self.arg_parser, and have it
            # raise an exception on error rather than printing to stderr
            # and exiting.
            self._cl_args = args or []

            def error(msg):
                raise ValueError(msg)

            if not from_cl:
                self.arg_parser.error = error

        self.load_args(self._cl_args)

        if (_im_func(self.load_options) !=
                _im_func(MRJobLauncher.load_options)):
            log.warning('load_options() is deprecated and will be'
                        ' removed in v0.7.0; please use load_args()'
                        ' instead.')
            self.load_options(self._cl_args)

        # Make it possible to redirect stdin, stdout, and stderr, for testing
        # See stdin, stdout, stderr properties and sandbox(), below.
        self._stdin = None
        self._stdout = None
        self._stderr = None

    # by default, self.stdin, self.stdout, and self.stderr are sys.std*.buffer
    # if it exists, and otherwise sys.std* otherwise (they should always deal
    # with bytes, not Unicode).
    #
    # *buffer* is pretty much a Python 3 thing, though some platforms
    # (notably Jupyterhub) don't have it. See #1441

    @property
    def stdin(self):
        return self._stdin or getattr(sys.stdin, 'buffer', sys.stdin)

    @property
    def stdout(self):
        return self._stdout or getattr(sys.stdout, 'buffer', sys.stdout)

    @property
    def stderr(self):
        return self._stderr or getattr(sys.stderr, 'buffer', sys.stderr)

    @classmethod
    def _usage(cls):
        """Command line usage string for this class"""
        return ("mrjob run [script path|executable path|--help]"
                " [options] [input files]")

    def _print_help(self, options):
        """Print help for this job. This will either print runner
        or basic help. Override to allow other kinds of help."""
        if options.runner:
            _print_help_for_runner(
                self._runner_opt_names_for_help(), options.deprecated)
        else:
            _print_basic_help(self.arg_parser,
                              self._usage(),
                              options.deprecated)

    @classmethod
    def run(cls, args=_READ_ARGS_FROM_SYS_ARGV):
        """Entry point for running job from the command-line.

        This is also the entry point when a mapper or reducer is run
        by Hadoop Streaming.

        Does one of:

        * Print step information (:option:`--steps`). See :py:meth:`show_steps`
        * Run a mapper (:option:`--mapper`). See :py:meth:`run_mapper`
        * Run a combiner (:option:`--combiner`). See :py:meth:`run_combiner`
        * Run a reducer (:option:`--reducer`). See :py:meth:`run_reducer`
        * Run the entire job. See :py:meth:`run_job`
        """
        # load options from the command line
        launcher = cls(args=args)
        launcher.run_job()

    def execute(self):
        # Launcher only runs jobs, doesn't do any Hadoop Streaming stuff
        self.run_job()

    def make_runner(self):
        """Make a runner based on command-line arguments, so we can
        launch this job on EMR, on Hadoop, or locally.

        :rtype: :py:class:`mrjob.runner.MRJobRunner`
        """
        return self._runner_class()(**self._runner_kwargs())

    @classmethod
    def set_up_logging(cls, quiet=False, verbose=False, stream=None):
        """Set up logging when running from the command line. This is also
        used by the various command-line utilities.

        :param bool quiet: If true, don't log. Overrides *verbose*.
        :param bool verbose: If true, set log level to ``DEBUG`` (default is
                             ``INFO``)
        :param bool stream: Stream to log to (default is ``sys.stderr``)
        """
        if quiet:
            log_to_null(name='mrjob')
            log_to_null(name='__main__')
        else:
            log_to_stream(name='mrjob', debug=verbose, stream=stream)
            log_to_stream(name='__main__', debug=verbose, stream=stream)

    def run_job(self):
        """Run the all steps of the job, logging errors (and debugging output
        if :option:`--verbose` is specified) to STDERR and streaming the
        output to STDOUT.

        Called from :py:meth:`run`. You'd probably only want to call this
        directly from automated tests.
        """
        # self.stderr is strictly binary, need to wrap it so it's possible
        # to log to it in Python 3
        log_stream = codecs.getwriter('utf_8')(self.stderr)

        self.set_up_logging(quiet=self.options.quiet,
                            verbose=self.options.verbose,
                            stream=log_stream)

        with self.make_runner() as runner:
            try:
                runner.run()
            except StepFailedException as e:
                # no need for a runner stacktrace if step failed; runners will
                # log more useful information anyway
                log.error(str(e))
                sys.exit(1)

            if self._should_cat_output():
                for chunk in runner.cat_output():
                    self.stdout.write(chunk)
                self.stdout.flush()

    def _should_cat_output(self):
        if self.options.cat_output is None:
            return not self.options.output_dir
        else:
            return self.options.cat_output

    ### Command-line arguments ###

    def configure_args(self):
        """Define arguments for this script. Called from :py:meth:`__init__()`.

        Re-define to define custom command-line arguments or pass
        through existing ones::

            def configure_args(self):
                super(MRYourJob, self).configure_args()

                self.add_passthru_arg(...)
                self.add_file_arg(...)
                self.pass_arg_through(...)
                ...
        """
        # if script path isn't set, expect it on the command line
        if self._FIRST_ARG_IS_SCRIPT_PATH:
            self.arg_parser.add_argument(
                dest='script_path',
                help='path of script to launch')

        self.arg_parser.add_argument(
            dest='args', nargs='*',
            help=('input paths to read (or stdin if not set). If --spark'
                  ' is set, the input and output path for the spark job.'))

        _add_basic_args(self.arg_parser)
        _add_job_args(self.arg_parser)
        _add_runner_args(self.arg_parser)

    def load_args(self, args):
        """Load command-line options into ``self.options`` and
        ``self._script_path``.

        Called from :py:meth:`__init__()` after :py:meth:`configure_args`.

        :type args: list of str
        :param args: a list of command line arguments. ``None`` will be
                     treated the same as ``[]``.

        Re-define if you want to post-process command-line arguments::

            def load_args(self, args):
                super(MRYourJob, self).load_args(args)

                self.stop_words = self.options.stop_words.split(',')
                ...
        """
        self.options = self.arg_parser.parse_args(args)

        if self.options.help:
            self._print_help(self.options)
            sys.exit(0)

        if self._FIRST_ARG_IS_SCRIPT_PATH:
            # should always be set, just hedging
            self._script_path = self.options.script_path

    def add_file_arg(self, *args, **kwargs):
        """Add a command-line option that sends an external file
        (e.g. a SQLite DB) to Hadoop::

             def configure_args(self):
                super(MRYourJob, self).configure_args()
                self.add_file_arg('--scoring-db', help=...)

        This does the right thing: the file will be uploaded to the working
        dir of the script on Hadoop, and the script will be passed the same
        option, but with the local name of the file in the script's working
        directory.

        .. note::

           If you pass a file to a job, best practice is to lazy-load its
           contents (e.g. make a method that opens the file the first time
           you call it) rather than loading it in your job's constructor or
           :py:meth:`load_args`. Not only is this more efficient, it's
           necessary if you want to run your job in a Spark executor
           (because the file may not be in the same place in a Spark driver).

        .. note::

           We suggest against sending Berkeley DBs to your job, as
           Berkeley DB is not forwards-compatible (so a Berkeley DB that you
           construct on your computer may not be readable from within
           Hadoop). Use SQLite databases instead. If all you need is an on-disk
           hash table, try out the :py:mod:`sqlite3dbm` module.

        .. versionchanged:: 0.6.6

           now accepts explicit ``type=str``

        .. versionchanged:: 0.6.8

           fully supported on Spark, including ``local[*]`` master
        """
        if kwargs.get('type') not in (None, str):
            raise ArgumentTypeError(
                'file options must take strings')

        if kwargs.get('action') not in (None, 'append', 'store'):
            raise ArgumentTypeError(
                "file options must use the actions 'store' or 'append'")

        pass_opt = self.arg_parser.add_argument(*args, **kwargs)

        self._file_arg_dests.add(pass_opt.dest)

    def add_passthru_arg(self, *args, **kwargs):
        """Function to create options which both the job runner
        and the job itself respect (we use this for protocols, for example).

        Use it like you would use
        :py:func:`argparse.ArgumentParser.add_argument`::

            def configure_args(self):
                super(MRYourJob, self).configure_args()
                self.add_passthru_arg(
                    '--max-ngram-size', type=int, default=4, help='...')

        If you want to pass files through to the mapper/reducer, use
        :py:meth:`add_file_arg` instead.

        If you want to pass through a built-in option (e.g. ``--runner``, use
        :py:meth:`pass_arg_through` instead.
        """
        pass_opt = self.arg_parser.add_argument(*args, **kwargs)

        self._passthru_arg_dests.add(pass_opt.dest)

    def pass_arg_through(self, opt_str):
        """Pass the given argument through to the job."""

        # _actions is hidden but the interface appears to be stable,
        # and there's no non-hidden interface we can use
        for action in self.arg_parser._actions:
            if opt_str in action.option_strings or opt_str == action.dest:
                self._passthru_arg_dests.add(action.dest)
                break
        else:
            raise ValueError('unknown arg: %s', opt_str)

    def is_task(self):
        """True if this is a mapper, combiner, or reducer.

        This is mostly useful inside :py:meth:`load_args`, to disable
        loading args when we aren't running inside Hadoop Streaming.
        """
        return False

    ### old optparse shims ###

    @property
    def args(self):
        class_name = self.__class__.__name__
        log.warning(
            '%s.args is a deprecated alias for %s.options.args, and will'
            ' be removed in v0.7.0' % (class_name, class_name))
        return self.options.args

    def configure_options(self):
        """.. deprecated:: 0.6.0

        Use `:py:meth:`configure_args` instead.
        """
        pass  # deprecation warning is in __init__()

    def load_options(self, args):
        """.. deprecated:: 0.6.0

        Use `:py:meth:`load_args` instead.
        """
        pass  # deprecation warning is in __init__()

    def add_file_option(self, *args, **kwargs):
        """.. deprecated:: 0.6.0

        Like :py:meth:`add_file_arg` except that it emulates the
        old :py:mod:`optparse` interface (which is almost identical).

        .. versionchanged:: 0.6.6

           accepts ``type='str'`` (used to only accept ``type='string'``)
        """
        log.warning(
            'add_file_option() is deprecated and will be removed in'
            ' v0.7.0. Use add_file_arg() instead.')

        self.add_file_arg(*args, **_optparse_kwargs_to_argparse(**kwargs))

    def add_passthrough_option(self, *args, **kwargs):
        """.. deprecated:: 0.6.0

        Like :py:meth:`add_passthru_arg` except that it emulates the
        old :py:mod:`optparse` interface (which is almost identical).

        .. versionchanged:: 0.6.6

           accepts ``type='str'`` (used to only accept ``type='string'``)
        """
        log.warning(
            'add_passthrough_option() is deprecated and will be removed in'
            ' v0.7.0. Use add_passthru_arg() instead.')

        self.add_passthru_arg(*args, **_optparse_kwargs_to_argparse(**kwargs))

    def pass_through_option(self, opt_str):
        """.. deprecated:: 0.6.0

        Like :py:meth:`pass_arg_through` except that it emulates the
        old :py:mod:`optparse` interface (which is almost identical).

        .. versionchanged:: 0.6.6

           accepts ``type='str'`` (used to only accept ``type='string'``)
        """
        log.warning(
            'pass_through_option() is deprecated and will be removed in'
            ' v0.7.0. Use pass_arg_through() instead.')

        self.pass_arg_through(opt_str)

    ### runners ###

    def _runner_class(self):
        """Runner class, as indicated by ``--runner``. This uses conditional
        imports to avoid importing runner modules that we don't need (and may
        not have libraries for).

        Defaults to ``'local'`` and disallows use of inline runner.
        """
        if not self.options.runner:
            return LocalMRJobRunner

        elif self.options.runner == 'inline':
            raise ValueError(
                "inline is not supported in the multi-lingual"
                " launcher.")

        return _runner_class(self.options.runner)

    def _runner_kwargs(self):
        # just use combine_dicts() and not combine_confs(); leave the
        # magic to the runner
        return combine_dicts(
            self._non_option_kwargs(),
            # don't screen out irrelevant opts (see #1898)
            self._kwargs_from_switches(set(_RUNNER_OPTS)),
            self._job_kwargs(),
        )

    def _runner_opt_names_for_help(self):
        opts = set(self._runner_class().OPT_NAMES)

        if self.options.runner == 'spark':
            # specific to Spark runner, but command-line only, so it doesn't
            # appear in SparkMRJobRunner.OPT_NAMES (see #2040)
            opts.add('max_output_files')

        return opts

    def _non_option_kwargs(self):
        """Keyword arguments to runner constructor that can't be set
        in mrjob.conf.

        These should match the (named) arguments to
        :py:meth:`~mrjob.runner.MRJobRunner.__init__`.
        """
        # build extra_args
        raw_args = _parse_raw_args(self.arg_parser, self._cl_args)

        extra_args = []

        for dest, option_string, args in raw_args:
            if dest in self._file_arg_dests:
                extra_args.append(option_string)
                extra_args.append(parse_legacy_hash_path('file', args[0]))
            elif dest in self._passthru_arg_dests:
                # special case for --hadoop-arg=-verbose etc.
                if (option_string and len(args) == 1 and
                        args[0].startswith('-')):
                    extra_args.append('%s=%s' % (option_string, args[0]))
                else:
                    if option_string:
                        extra_args.append(option_string)
                    extra_args.extend(args)

        # max_output_files is added by _add_runner_args() but can only
        # be set from the command line, so we add it here (see #2040)
        return dict(
            conf_paths=self.options.conf_paths,
            extra_args=extra_args,
            hadoop_input_format=self.hadoop_input_format(),
            hadoop_output_format=self.hadoop_output_format(),
            input_paths=self.options.args,
            max_output_files=self.options.max_output_files,
            mr_job_script=self._script_path,
            output_dir=self.options.output_dir,
            partitioner=self.partitioner(),
            stdin=self.stdin,
            step_output_dir=self.options.step_output_dir,
        )

    def _kwargs_from_switches(self, keys):
        return dict(
            (key, getattr(self.options, key))
            for key in keys if hasattr(self.options, key)
        )

    def _job_kwargs(self):
        """Keyword arguments to the runner class that can be specified
        by the job/launcher itself."""
        # use the most basic combiners; leave magic like resolving paths
        # and blanking out jobconf values to the runner
        return dict(
            # command-line has the final say on jobconf and libjars
            jobconf=combine_dicts(
                self.jobconf(), self.options.jobconf),
            libjars=combine_lists(
                self.libjars(), self.options.libjars),
            partitioner=self.partitioner(),
            sort_values=self.sort_values(),
            # TODO: should probably put self.options last below for consistency
            upload_archives=combine_lists(
                self.options.upload_archives, self.archives()),
            upload_dirs=combine_lists(
                self.options.upload_dirs, self.dirs()),
            upload_files=combine_lists(
                self.options.upload_files, self.files()),
        )

    ### Hooks for options defined by the job ###

    def archives(self):
        """See :py:meth:`mrjob.job.MRJob.files`."""
        return []

    def dirs(self):
        """See :py:meth:`mrjob.job.MRJob.dirs`."""
        return []

    def files(self):
        """See :py:meth:`mrjob.job.MRJob.files`."""
        return []

    def hadoop_input_format(self):
        """See :py:meth:`mrjob.job.MRJob.hadoop_input_format`."""
        return None

    def hadoop_output_format(self):
        """See :py:meth:`mrjob.job.MRJob.hadoop_output_format`."""
        return None

    def jobconf(self):
        """See :py:meth:`mrjob.job.MRJob.jobconf`."""
        return {}

    def libjars(self):
        """See :py:meth:`mrjob.job.MRJob.libjars`."""
        return []

    def partitioner(self):
        """See :py:meth:`mrjob.job.MRJob.partitioner`."""
        return None

    def sort_values(self):
        """See :py:meth:`mrjob.job.MRJob.sort_values`."""
        return None

    ### Testing ###

    def sandbox(self, stdin=None, stdout=None, stderr=None):
        """Redirect stdin, stdout, and stderr for automated testing.

        You can set stdin, stdout, and stderr to file objects. By
        default, they'll be set to empty ``BytesIO`` objects.
        You can then access the job's file handles through ``self.stdin``,
        ``self.stdout``, and ``self.stderr``. See :ref:`testing` for more
        information about testing.

        You may call sandbox multiple times (this will essentially clear
        the file handles).

        ``stdin`` is empty by default. You can set it to anything that yields
        lines::

            mr_job.sandbox(stdin=BytesIO(b'some_data\\n'))

        or, equivalently::

            mr_job.sandbox(stdin=[b'some_data\\n'])

        For convenience, this sandbox() returns self, so you can do::

            mr_job = MRJobClassToTest().sandbox()

        Simple testing example::

            mr_job = MRYourJob.sandbox()
            self.assertEqual(list(mr_job.reducer('foo', ['a', 'b'])), [...])

        More complex testing example::

            from BytesIO import BytesIO

            from mrjob.parse import parse_mr_job_stderr
            from mrjob.protocol import JSONProtocol

            mr_job = MRYourJob(args=[...])

            fake_input = '"foo"\\t"bar"\\n"foo"\\t"baz"\\n'
            mr_job.sandbox(stdin=BytesIO(fake_input))

            mr_job.run_reducer(link_num=0)

            self.assertEqual(mrjob.stdout.getvalue(), ...)
            self.assertEqual(parse_mr_job_stderr(mr_job.stderr), ...)

        .. note::

           If you are using Spark, it's recommended you only pass in
           :py:class:`io.BytesIO` or other serializable alternatives to file
           objects. *stdin*, *stdout*, and *stderr* get stored as job
           attributes, which means if they aren't serializable, neither
           is the job instance or its methods.
        """
        self._stdin = stdin or BytesIO()
        self._stdout = stdout or BytesIO()
        self._stderr = stderr or BytesIO()

        return self


if __name__ == '__main__':
    MRJobLauncher.run()



# Copyright 2018 Yelp
# Copyright 2019 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test the spark-submit script."""
from __future__ import print_function

import os
from glob import glob
from os.path import join
from os.path import exists
from unittest import skipIf

try:
    import pyspark
except ImportError:
    pyspark = None

try:
    from mrjob.examples import spark_wordcount_script
except ImportError:
    spark_wordcount_script = None

from mrjob.runner import _runner_class
from mrjob.step import StepFailedException
from mrjob.tools.spark_submit import main as spark_submit_main
from mrjob.util import safeeval

from tests.mock_boto3.case import MockBoto3TestCase
from tests.py2 import MagicMock
from tests.py2 import Mock
from tests.py2 import patch
from tests.sandbox import SandboxedTestCase


class MockSystemExit(Exception):
    pass


class SparkSubmitToolTestCase(SandboxedTestCase):

    def setUp(self):
        super(SparkSubmitToolTestCase, self).setUp()

        self.runner_class = None

        self.runner = MagicMock()

        def _mock_runner_class(runner_alias):
            rc = _runner_class(runner_alias)

            self.runner_class = Mock()
            self.runner_class.return_value = self.runner
            self.runner_class.alias = rc.alias
            self.runner_class.OPT_NAMES = rc.OPT_NAMES

            return self.runner_class

        self.runner_class = self.start(patch(
            'mrjob.tools.spark_submit._runner_class',
            side_effect=_mock_runner_class))

        self.runner_log = self.start(patch('mrjob.runner.log'))

        # don't actually want to exit after printing help
        self.exit = self.start(patch('sys.exit', side_effect=MockSystemExit))

        # don't set up logging
        self.set_up_logging = self.start(
            patch('mrjob.job.MRJob.set_up_logging'))

        # save printout, rather than actually printing
        self.printout = ''

        def _mock_print_message(self_, message, file=None):
            self.printout += message

        self.start(patch('argparse.ArgumentParser._print_message',
                         _mock_print_message))

        def _mock_print(s=''):
            self.printout += s + '\n'

        # print() isn't considered part of the module in Python 3.4
        self.start(patch('mrjob.tools.spark_submit.print',
                         _mock_print, create=True))

    def get_runner_kwargs(self):
        return self.runner_class.call_args_list[-1][1]

    def test_basic(self):
        spark_submit_main(['foo.py', 'arg1', 'arg2'])

        self.assertEqual(self.runner_class.alias, 'spark')

        self.assertTrue(self.runner_class.called)
        self.assertTrue(self.runner.run.called)

        kwargs = self.get_runner_kwargs()

        self.assertEqual(kwargs['steps'], [
            dict(
                args=['arg1', 'arg2'],
                jobconf={},
                script='foo.py',
                spark_args=[],
                type='spark_script',
            )
        ])

    def test_switches_to_spark_script(self):
        # regression test for #2070
        spark_submit_main(['foo.py', '--bar', 'baz'])

        self.assertEqual(self.runner_class.alias, 'spark')

        self.assertTrue(self.runner_class.called)
        self.assertTrue(self.runner.run.called)

        kwargs = self.get_runner_kwargs()

        self.assertEqual(kwargs['steps'], [
            dict(
                args=['--bar', 'baz'],
                jobconf={},
                script='foo.py',
                spark_args=[],
                type='spark_script',
            )
        ])

    def test_jar_step(self):
        spark_submit_main(['dora.jar', 'arg1', 'arg2', 'arg3'])

        self.assertEqual(self.runner_class.alias, 'spark')

        self.assertTrue(self.runner_class.called)
        self.assertTrue(self.runner.run.called)

        kwargs = self.get_runner_kwargs()

        self.assertEqual(kwargs['steps'], [
            dict(
                args=['arg1', 'arg2', 'arg3'],
                jar='dora.jar',
                jobconf={},
                main_class=None,
                spark_args=[],
                type='spark_jar',
            )
        ])

    def test_jar_main_class(self):
        spark_submit_main(['--class', 'Backpack',
                           'dora.jar', 'arg1', 'arg2', 'arg3'])

        kwargs = self.get_runner_kwargs()

        self.assertEqual(kwargs['steps'], [
            dict(
                args=['arg1', 'arg2', 'arg3'],
                jar='dora.jar',
                jobconf={},
                main_class='Backpack',
                spark_args=[],
                type='spark_jar',
            )
        ])

    def test_set_runner_class(self):
        spark_submit_main(['-r', 'emr', 'foo.py', 'arg1'])

        self.assertEqual(self.runner_class.alias, 'emr')

        self.assertTrue(self.runner_class.called)
        self.assertTrue(self.runner.run.called)

    def test_no_script_args_okay(self):
        spark_submit_main(['foo.py'])

        kwargs = self.get_runner_kwargs()

        self.assertEqual(kwargs['steps'], [
            dict(
                args=[],
                jobconf={},
                script='foo.py',
                spark_args=[],
                type='spark_script',
            )
        ])

    def test_allow_py3_extension(self):
        spark_submit_main(['foo.py3', 'arg1', 'arg2'])

        kwargs = self.get_runner_kwargs()

        self.assertEqual(kwargs['steps'], [
            dict(
                args=['arg1', 'arg2'],
                jobconf={},
                script='foo.py3',
                spark_args=[],
                type='spark_script',
            )
        ])

    def test_not_py_or_jar(self):
        self.assertRaises(ValueError, spark_submit_main,
                          ['whoo.sh', 'arg1'])

    def test_pass_through_to_step_spark_args(self):
        spark_submit_main(['--class', 'Backpack',
                           '--name', 'Backpack',
                           '--num-executors', '3',
                           '--conf', 'foo=BAR',
                           '--name', 'Mochila',
                           'dora.jar', 'arg1'])

        # --class becomes part of step
        # --conf is an alias for a mrjob opt, goes to runner
        # other args end up in spark-args as-is
        kwargs = self.get_runner_kwargs()

        self.assertEqual(kwargs['steps'], [
            dict(
                args=['arg1'],
                jar='dora.jar',
                jobconf={},
                main_class='Backpack',
                spark_args=[
                    '--name', 'Backpack',
                    '--num-executors', '3',
                    '--name', 'Mochila',
                ],
                type='spark_jar',
            )
        ])

        self.assertEqual(kwargs['jobconf'], dict(foo='BAR'))

    def test_runner_kwargs(self):
        spark_submit_main(['--hadoop-bin', 'super-hadoop',
                           '--master', 'local',
                           '--py-files', 'bar.py,baz.py',
                           'foo.py', 'arg1'])

        kwargs = self.get_runner_kwargs()

        # regular old runner arg
        self.assertEqual(kwargs['hadoop_bin'], 'super-hadoop')

        # spark alias for mrjob opt
        self.assertEqual(kwargs['spark_master'], 'local')

        # arg with custom parser
        self.assertEqual(kwargs['py_files'], ['bar.py', 'baz.py'])

    def test_filters_runner_kwargs(self):
        # may want to change this behavior; see #1898
        spark_submit_main(['-r', 'emr', 'foo.py', 'arg1'])

        kwargs = self.get_runner_kwargs()

        self.assertIn('region', kwargs)
        self.assertNotIn('hadoop_bin', kwargs)

    def test_cleanup_called(self):
        spark_submit_main(['-r', 'emr', 'foo.py', 'arg1'])
        self.assertTrue(self.runner.cleanup.called)

    def test_cleanup_called_even_if_run_failed(self):
        self.runner.run.side_effect = StepFailedException

        self.assertRaises(StepFailedException, spark_submit_main,
                          ['-r', 'emr', 'foo.py', 'arg1'])
        self.assertTrue(self.runner.cleanup.called)

    def test_hard_coded_kwargs(self):
        spark_submit_main(['foo.py', 'arg1'])

        kwargs = self.get_runner_kwargs()

        self.assertEqual(kwargs['check_input_paths'], False)
        self.assertEqual(kwargs['input_paths'], [os.devnull])
        self.assertEqual(kwargs['output_dir'], None)

    def test_no_switches_for_hard_coded_kwargs(self):
        self.assertRaises(MockSystemExit, spark_submit_main,
                          ['--check-input-paths', 'foo.py', 'arg1'])
        self.assertRaises(MockSystemExit, spark_submit_main,
                          ['--output-dir', 'out', 'foo.py', 'arg1'])

    def _assert_prints_basic_help(self, args, include_deprecated=False):
        self.assertRaises(MockSystemExit, spark_submit_main, args)
        self.exit.assert_called_once_with(0)

        self.assertTrue(self.printout.startswith('usage:'))

        # has spark switches
        self.assertIn('--jars', self.printout)
        self.assertIn('--name', self.printout)
        self.assertIn('--py-files', self.printout)
        # but not runner switches or aliases
        self.assertNotIn('--python-bin', self.printout)
        self.assertNotIn('--libjars', self.printout)

        # --deprecated appears either way; either it appears as
        # a switch, or as a message at the end explaining how
        # to use it with --help
        if include_deprecated:
            self.assertIn('--deprecated', self.printout[:-50])
        else:
            self.assertIn('--deprecated', self.printout[-50:])

    def _assert_prints_runner_help(self, args, runner_alias,
                                   expect_switches=(),
                                   dont_expect_switches=()):
        self.assertRaises(MockSystemExit, spark_submit_main, args)
        self.exit.assert_called_once_with(0)

        # doesn't show usage, should include runner name in first line
        first_line = self.printout.split('\n')[0]
        self.assertIn('optional arguments', first_line)
        self.assertIn(runner_alias, first_line)

        # has runner switches
        self.assertIn('--python-bin', self.printout)
        self.assertIn('--libjars', self.printout)

        # but not basic spark switches
        self.assertNotIn('--name', self.printout)

        for switch in expect_switches:
            self.assertIn(switch, self.printout)

        for switch in dont_expect_switches:
            self.assertNotIn(switch, self.printout)

    def test_help_arg(self):
        self._assert_prints_basic_help(['-h'])

    def test_help_arg_with_deprecated(self):
        self._assert_prints_basic_help(['-h', '--deprecated'],
                                       include_deprecated=True)

    def test_help_arg_with_runner(self):
        self._assert_prints_runner_help(
            ['-h', '-r', 'emr'],
            'emr',
            expect_switches=['--region'],
            # --hadoop-bin is not for emr, --py-file is deprecated
            dont_expect_switches=['--hadoop-bin', '--py-file'])

    def test_help_arg_with_runner_and_deprecated(self):
        self._assert_prints_runner_help(
            ['-h', '-r', 'emr', '--deprecated'],
            'emr',
            expect_switches=['--region', '--py-file'],
            # we don't actually mention --deprecated in runner help
            dont_expect_switches=['--hadoop-bin', '--deprecated'])


class SparkSubmitToEMRTestCase(MockBoto3TestCase):

    def setUp(self):
        super(SparkSubmitToEMRTestCase, self).setUp()

        # don't set up logging
        self.set_up_logging = self.start(
            patch('mrjob.job.MRJob.set_up_logging'))

    def test_end_to_end(self):
        script_path = self.makefile('foo.py')

        spark_submit_main(
            ['-r', 'emr', script_path, 'arg1'])

        emr_client = self.client('emr')

        cluster_ids = [c['Id'] for c in
                       emr_client.list_clusters()['Clusters']]
        self.assertEqual(len(cluster_ids), 1)
        cluster_id = cluster_ids[0]

        steps = emr_client.list_steps(ClusterId=cluster_id)['Steps']
        self.assertEqual(len(steps), 1)
        step = steps[0]

        self.assertEqual(step['Status']['State'], 'COMPLETED')
        step_args = step['Config']['Args']

        self.assertEqual(step_args[0], 'spark-submit')
        self.assertEqual(step_args[-1], 'arg1')
        self.assertTrue(step_args[-2].endswith('/foo.py'))


@skipIf(pyspark is None, 'no pyspark module')
class SparkSubmitLocallyTestCase(SandboxedTestCase):
    # this test is a bit slow (~30s) as it actually runs on Spark
    # in local-cluster mode

    def setUp(self):
        super(SparkSubmitLocallyTestCase, self).setUp()

        # don't set up logging
        self.set_up_logging = self.start(
            patch('mrjob.job.MRJob.set_up_logging'))

    def test_end_to_end(self):
        script_path = spark_wordcount_script.__file__
        if script_path.endswith('.pyc'):
            script_path = script_path[:-1]

        input_path = self.makefile(
            'input', b'one fish\ntwo fish\nred fish\nblue fish\n')

        # don't create this path, let Spark do it
        output_path = join(self.tmp_dir, 'output')
        self.assertFalse(exists(output_path))

        spark_submit_main(
            ['-r', 'local', script_path, input_path, output_path])

        self.assertTrue(exists(output_path))

        word_counts = {}

        for path in glob(join(output_path, 'part-*')):
            with open(path) as f:
                for line in f:
                    word, count = safeeval(line)
                    word_counts[word] = count

        self.assertEqual(word_counts,
                         dict(blue=1, fish=4, one=1, red=1, two=1))



# -*- coding: utf-8 -*-
# Copyright 2009-2013 Yelp and Contributors
# Copyright 2015-2017 Yelp
# Copyright 2018 Yelp and Contributors
# Copyright 2019 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run an MRJob locally by forking off a bunch of processes and piping
them together. Useful for testing, not terrible for running medium-sized
jobs on all CPUs."""
import logging
import math
import os
import platform
from functools import partial
from multiprocessing import Pool
from subprocess import CalledProcessError
from subprocess import check_call

try:
    import pty
    pty  # quiet "redefinition of unused ..." warning from pyflakes
except ImportError:
    pty = None

from mrjob.bin import MRJobBinRunner
from mrjob.logs.errors import _format_error
from mrjob.logs.step import _log_log4j_record
from mrjob.logs.task import _parse_task_stderr
from mrjob.py2 import string_types
from mrjob.sim import SimMRJobRunner
from mrjob.sim import _sort_lines_in_memory
from mrjob.step import StepFailedException
from mrjob.util import cmd_line

log = logging.getLogger(__name__)


_DEFAULT_EXECUTOR_MEMORY = '1g'


class _TaskFailedException(StepFailedException):
    """Extension of :py:class:`~mrjob.step.StepFailedException` that
    blames one particular task."""
    _FIELDS = StepFailedException._FIELDS + ('task_type', 'task_num')

    def __init__(
            self, reason=None, step_num=None, num_steps=None, step_desc=None,
            task_type=None, task_num=None):
        super(_TaskFailedException, self).__init__(
            reason=reason, step_num=step_num,
            num_steps=num_steps, step_desc=step_desc)

        self.task_type = task_type
        self.task_num = task_num


class LocalMRJobRunner(SimMRJobRunner, MRJobBinRunner):
    """Runs an :py:class:`~mrjob.job.MRJob` locally, for testing purposes.
    Invoked when you run your job with ``-r local``.

    Unlike :py:class:`~mrjob.job.InlineMRJobRunner`, this actually spawns
    multiple subprocesses for each task.

    It's rare to need to instantiate this class directly (see
    :py:meth:`~LocalMRJobRunner.__init__` for details).

    .. versionadded:: 0.6.8

       can run Spark steps as well, on the ``local-cluster`` Spark master.
    """
    alias = 'local'

    OPT_NAMES = SimMRJobRunner.OPT_NAMES | MRJobBinRunner.OPT_NAMES | {
        'sort_bin',
    }

    _STEP_TYPES = (
        SimMRJobRunner._STEP_TYPES | {'spark_jar', 'spark_script'})

    def __init__(self, **kwargs):
        """Arguments to this constructor may also appear in :file:`mrjob.conf`
        under ``runners/local``.

        :py:class:`~mrjob.local.LocalMRJobRunner`'s constructor takes the
        same keyword args as
        :py:class:`~mrjob.runner.MRJobRunner`. However, please note:

        * *cmdenv* is combined with :py:func:`~mrjob.conf.combine_local_envs`
        * *python_bin* defaults to ``sys.executable`` (the current python
          interpreter)
        * *hadoop_input_format*, *hadoop_output_format*,
          and *partitioner* are ignored because they
          require Java. If you need to test these, consider starting up a
          standalone Hadoop instance and running your job with ``-r hadoop``.
        """
        super(LocalMRJobRunner, self).__init__(**kwargs)

    def _invoke_task_func(self, task_type, step_num, task_num):
        args = self._substep_args(step_num, task_type)
        num_steps = self._num_steps()

        # stdin, stdout, stderr, wd, and env will be passed in later
        return partial(
            _invoke_task_in_subprocess,
            task_type, step_num, task_num,
            args, num_steps)

    def _run_step_on_spark(self, step, step_num):
        if self._opts['upload_archives']:
            log.warning('Spark master %r will probably ignore archives' %
                        self._spark_master())

        spark_submit_args = self._args_for_spark_step(step_num)

        env = dict(os.environ)
        env.update(self._spark_cmdenv(step_num))

        returncode = self._run_spark_submit(spark_submit_args, env,
                                            record_callback=_log_log4j_record)

        if returncode:
            reason = str(CalledProcessError(returncode, spark_submit_args))
            raise StepFailedException(
                reason=reason, step_num=step_num,
                num_steps=self._num_steps())

    def _run_multiple(self, funcs, num_processes=None):
        """Use multiprocessing to run in parallel."""
        pool = Pool(processes=self._opts['num_cores'])

        try:
            results = [
                pool.apply_async(partial(_pickle_safe, func))
                for func in funcs
            ]

            for result in results:
                result.get()

        # make sure that the pool (and its file descriptors, etc.)
        # don't stay open. This doesn't matter much for individual jobs,
        # but it makes our automated tasks run out of file descriptors.

            pool.close()
        except:
            # if there's an error in one task, terminate all others
            pool.terminate()
            raise
        finally:
            pool.join()

    def _log_cause_of_error(self, ex):
        if not isinstance(ex, _TaskFailedException):
            # if something went wrong inside mrjob, the stacktrace
            # will bubble up to the top level
            return

        # not using LogInterpretationMixin because it would be overkill

        if not self._opts['read_logs']:
            return

        input_path = self._task_input_path(
            ex.task_type, ex.step_num, ex.task_num)
        stderr_path = self._task_stderr_path(
            ex.task_type, ex.step_num, ex.task_num)

        if self.fs.exists(stderr_path):  # it should, but just to be safe
            # log-parsing code expects "str", not bytes; open in text mode
            with open(stderr_path) as stderr:
                task_error = _parse_task_stderr(stderr)
                if task_error:
                    task_error['path'] = stderr_path
                    log.error('Cause of failure:\n\n%s\n\n' %
                              _format_error(dict(
                                  split=dict(path=input_path),
                                  task_error=task_error)))
                    return

        # fallback if we can't find the error (e.g. the job does something
        # weird to stderr or stack traces)
        log.error('Error while reading from %s:\n' % input_path)

    def _default_python_bin(self, local=False):
        """Always return *sys.executable*, if defined"""
        return super(LocalMRJobRunner, self)._default_python_bin(
            local=True)

    def _sort_input_func(self):
        """Try sorting with the :command:`sort` binary before falling
        back to in-memory sort."""
        if platform.system() == 'Windows':  # we assume Unix sort
            return super(LocalMRJobRunner, self)._sort_input_func()
        else:
            return partial(
                _sort_lines_with_sort_bin,
                sort_bin=self._sort_bin(),
                tmp_dir=self._get_local_tmp_dir())

    def _sort_bin(self):
        """The binary to use to sort input.

        (On Windows, we go straight to sorting in memory.)
        """
        if self._opts['sort_bin']:
            return self._opts['sort_bin']
        elif self._sort_values:
            return ['sort']
        else:
            # only sort on the reducer key (see #660)
            return ['sort', '-t', '\t', '-k', '1,1', '-s']

    # Spark steps

    # TODO: _spark_master() should probably take step_num, to allow for
    # step-specific jobconf
    def _spark_master(self):
        """Use the local-cluster master, which simulates a Spark cluster."""
        # figure out the required parameters to local-cluster
        num_executors = self._num_cores()

        # for now assigning one core per executor, so we don't have to worry
        # about a number of cores that's not evenly divisible
        cores_per_executor = 1

        executor_mem_bytes = _to_num_bytes(
            self._opts['jobconf'].get('spark.executor.memory') or
            _DEFAULT_EXECUTOR_MEMORY)
        executor_mem_mb = math.ceil(executor_mem_bytes / 1024.0 / 1024.0)

        return 'local-cluster[%d,%d,%d]' % (
            num_executors, cores_per_executor, executor_mem_mb)


def _to_num_bytes(java_mem_str):
    if isinstance(java_mem_str, string_types):
        for i, magnitude in enumerate(('k', 'm', 'g', 't'), start=1):
            if java_mem_str.lower().endswith(magnitude):
                return int(java_mem_str[:-1]) * 1024 ** i

    return int(java_mem_str)


# pickle utilities, to protect multiprocessing from itself

def _invoke_task_in_subprocess(
        task_type, step_num, task_num,
        args, num_steps,
        stdin, stdout, stderr, wd, env):
    """A pickleable function that invokes a task in a subprocess."""
    log.debug('> %s' % cmd_line(args))

    try:
        check_call(args, stdin=stdin, stdout=stdout, stderr=stderr,
                   cwd=wd, env=env)
    except Exception as ex:
        raise _TaskFailedException(
            reason=str(ex),
            step_num=step_num,
            num_steps=num_steps,
            task_type=task_type,
            task_num=task_num,
        )


def _pickle_safe(func):
    """Call no-args function *func*, returning *None* and ensuring
    that any exception raised is pickleable."""
    try:
        func()  # always return None
    except _TaskFailedException:
        raise  # we know these are pickleable
    except Exception as ex:
        raise Exception(repr(ex))  # we know this is pickleable


# other utilities

def _sort_lines_with_sort_bin(input_paths, output_path, sort_bin,
                              sort_values=False, tmp_dir=None):
    """Sort lines the given *input_paths* into *output_path*,
    using *sort_bin*. If there is a problem, fall back to in-memory sort.

    This is a helper for :py:meth:`LocalMRJobRunner._sort_input_func`.

    *tmp_dir* determines the value of :envvar:`$TMP` and :envvar:`$TMPDIR`
    that *sort_bin* sees.
    """
    if input_paths:
        env = os.environ.copy()

        # ignore locale when sorting
        env['LC_ALL'] = 'C'

        # Make sure that the tmp dir environment variables are changed if
        # the default is changed.
        env['TMP'] = tmp_dir
        env['TMPDIR'] = tmp_dir

        with open(output_path, 'wb') as output:
            args = sort_bin + list(input_paths)
            log.debug('> %s' % cmd_line(args))

            try:
                check_call(args, stdout=output, env=env)
                return
            except CalledProcessError:
                log.error(
                    '`%s` failed, falling back to in-memory sort' %
                    cmd_line(sort_bin))
            except OSError:
                log.error(
                    'no sort binary, falling back to in-memory sort')

    _sort_lines_in_memory(input_paths, output_path, sort_values=sort_values)



# Copyright 2015 Yelp
# Copyright 2017 Yelp
# Copyright 2019 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compatilibity layer for test code, mostly to deal with the fact
that mock is a library in Python 2 but built into Python 3.

Also provides utility code for mocking ``sys.stdout`` and ``sys.stderr``.
"""
import codecs
import sys
from io import BytesIO

from mrjob.py2 import PY2
from mrjob.py2 import StringIO

# mock is built into unittest in Python 3.3+
if sys.version_info < (3, 3):
    import mock
else:
    from unittest import mock

ANY = mock.ANY
ANY

MagicMock = mock.MagicMock
MagicMock

Mock = mock.Mock
Mock

call = mock.call
call

patch = mock.patch
patch

if PY2:
    ResourceWarning = Warning
else:
    ResourceWarning = ResourceWarning


def mock_stdout_or_stderr():
    """Use this to make a mock stdout/stderr (e.g. for a method
    that calls print()).

    In Python 2, this is just a StringIO.

    In Python 3, this is a BytesIO wrapped in a UTF-8 codec writer. You can
    access the BytesIO through the "buffer" attribute.

    In either case, you can get the mock output (as bytes) by calling
    getvalue().
    """
    if PY2:
        return StringIO()

    buf = BytesIO()
    writer = codecs.getwriter('utf_8')(buf)
    writer.buffer = buf
    writer.getvalue = lambda: buf.getvalue()

    return writer



# Copyright 2016 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Job to test if runners respect SORT_VALUES."""
from mrjob.job import MRJob
from mrjob.step import MRStep


class MRSortValuesAndMore(MRJob):

    JOBCONF = {
        'mapreduce.partition.keycomparator.options': '-k1 -k2nr',
        'stream.num.map.output.key.fields': 3,
    }

    PARTITIONER = 'org.apache.hadoop.mapred.lib.HashPartitioner'

    SORT_VALUES = True

    def dummy_mapper_init(self):
        pass

    def steps(self):
        return [
            MRStep(
                mapper_init=self.dummy_mapper_init,
            ),
            MRStep(
                mapper_init=self.dummy_mapper_init,
                jobconf={
                    'mapreduce.job.output.key.comparator.class':
                    'org.apache.hadoop.mapred.lib.KeyFieldBasedComparator',
                }
            ),
        ]


if __name__ == '__main__':
    MRSortValuesAndMore.run()



# Copyright 2009-2012 Yelp
# Copyright 2015-2018 Yelp
# Copyright 2019 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run a command on every node of a cluster. Store stdout and stderr for
results in OUTPUT_DIR.

Usage::

    mrjob boss CLUSTER_ID [options] "command string"

Options::

  -c CONF_PATHS, --conf-path CONF_PATHS
                        Path to alternate mrjob.conf file to read from
  --no-conf             Don't load mrjob.conf even if it's available
  --ec2-key-pair-file EC2_KEY_PAIR_FILE
                        Path to file containing SSH key for EMR
  --emr-endpoint EMR_ENDPOINT
                        Force mrjob to connect to EMR on this endpoint (e.g.
                        us-west-1.elasticmapreduce.amazonaws.com). Default is
                        to infer this from region.
  -h, --help            show this help message and exit
  -o OUTPUT_DIR, --output-dir OUTPUT_DIR
                        Specify an output directory (default: CLUSTER_ID)
  -q, --quiet           Don't print anything to stderr
  --region REGION       GCE/AWS region to run Dataproc/EMR jobs in.
  --s3-endpoint S3_ENDPOINT
                        Force mrjob to connect to S3 on this endpoint (e.g. s3
                        -us-west-1.amazonaws.com). You usually shouldn't set
                        this; by default mrjob will choose the correct
                        endpoint for each S3 bucket based on its location.
  --ssh-bin SSH_BIN     Name/path of ssh binary. Arguments are allowed (e.g.
                        --ssh-bin 'ssh -v')
  -v, --verbose         print more messages to stderr
"""
from __future__ import print_function

import os
from argparse import ArgumentParser

from mrjob.emr import EMRJobRunner
from mrjob.job import MRJob
from mrjob.options import _add_basic_args
from mrjob.options import _add_runner_args
from mrjob.options import _alphabetize_actions
from mrjob.options import _filter_by_role
from mrjob.py2 import to_unicode
from mrjob.util import shlex_split


def main(cl_args=None):
    usage = 'usage: %(prog)s boss CLUSTER_ID [options] "command string"'
    description = ('Run a command on the master and all worker nodes of an EMR'
                   ' cluster. Store stdout/stderr for results in OUTPUT_DIR.')

    arg_parser = ArgumentParser(usage=usage, description=description)
    arg_parser.add_argument('-o', '--output-dir', dest='output_dir',
                            default=None,
                            help="Specify an output directory (default:"
                            " CLUSTER_ID)")

    arg_parser.add_argument(dest='cluster_id',
                            help='ID of cluster to run command on')
    arg_parser.add_argument(dest='cmd_string',
                            help='command to run, as a single string')

    _add_basic_args(arg_parser)
    _add_runner_args(
        arg_parser,
        {'ec2_key_pair_file', 'ssh_bin'} | _filter_by_role(
            EMRJobRunner.OPT_NAMES, 'connect')
    )

    _alphabetize_actions(arg_parser)

    options = arg_parser.parse_args(cl_args)

    MRJob.set_up_logging(quiet=options.quiet, verbose=options.verbose)

    runner_kwargs = options.__dict__.copy()
    for unused_arg in ('cluster_id', 'cmd_string', 'output_dir',
                       'quiet', 'verbose'):
        del runner_kwargs[unused_arg]

    cmd_args = shlex_split(options.cmd_string)

    output_dir = os.path.abspath(options.output_dir or options.cluster_id)

    with EMRJobRunner(
            cluster_id=options.cluster_id, **runner_kwargs) as runner:
        _run_on_all_nodes(runner, output_dir, cmd_args)


def _run_on_all_nodes(runner, output_dir, cmd_args, print_stderr=True):
    """Given an :py:class:`EMRJobRunner`, run the command specified by
    *cmd_args* on all nodes in the cluster and save the stdout and stderr of
    each run to subdirectories of *output_dir*.
    """
    master_addr = runner._address_of_master()
    addresses = [master_addr]

    worker_addrs = runner._ssh_worker_hosts()

    if worker_addrs:
        addresses += ['%s!%s' % (master_addr, worker_addr)
                      for worker_addr in worker_addrs]

    for addr in addresses:

        stdout, stderr = runner.fs.ssh._ssh_run(addr, cmd_args)

        if print_stderr:
            print('---')
            print('Command completed on %s.' % addr)
            print(to_unicode(stderr), end=' ')

        if '!' in addr:
            base_dir = os.path.join(output_dir, 'worker ' + addr.split('!')[1])
        else:
            base_dir = os.path.join(output_dir, 'master')

        if not os.path.exists(base_dir):
            os.makedirs(base_dir)

        with open(os.path.join(base_dir, 'stdout'), 'wb') as f:
            f.write(stdout)

        with open(os.path.join(base_dir, 'stderr'), 'wb') as f:
            f.write(stderr)


if __name__ == '__main__':
    main()



# Copyright 2017 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A job that follows the MRJob protocol but doesn't inherit from mrjob.

This counts the number of "words" in a document
"""
import json
import sys

_STEPS = [
    dict(
        type='streaming',
        mapper=dict(type='script'),
        reducer=dict(type='script'),
    ),
]


def main():
    args = sys.argv[1:]

    if '--steps' in args:
        print(json.dumps(_STEPS))
    elif '--mapper' in args:
        for line in sys.stdin:
            print('words\t%d' % len(line.split()))
    elif '--reducer' in args:
        num_words = sum(int(line.split()[-1]) for line in sys.stdin)
        if num_words:  # account for reducer that's passed an empty file
            print(num_words)


if __name__ == '__main__':
    main()



# Copyright 2013 David Marin
# Copyright 2016 Yelp
# Copyright 2018 Yelp and Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A simple example of linking a hadoop jar with a Python step. This
calculates the frequency of various word frequencies in a text file.

This example works out-of-the box on EMR and Google Cloud Dataproc.

This also only works on a single input path/directory, due to limitations
of the example jar.
"""
from mrjob.job import MRJob
from mrjob.protocol import RawProtocol
from mrjob.step import GENERIC_ARGS
from mrjob.step import INPUT
from mrjob.step import JarStep
from mrjob.step import MRStep
from mrjob.step import OUTPUT

# use the file:// trick to access a jar hosted on the cloud
_EXAMPLES_JAR_URI = (
    'file:///usr/lib/hadoop-mapreduce/hadoop-mapreduce-examples.jar')

_EMR_HADOOP_PRE_4_X_JAR_URI = (
    'file:///home/hadoop/hadoop-examples.jar')

_WORDCOUNT_MAIN_CLASS = 'org.apache.hadoop.examples.WordCount'


class MRJarStepExample(MRJob):
    """A contrived example that runs wordcount from the hadoop example
    jar, and then does a frequency count of the frequencies."""

    def configure_args(self):
        super(MRJarStepExample, self).configure_args()

        self.add_passthru_arg(
            '--use-main-class', dest='use_main_class',
            default=False, action='store_true')

        self.add_passthru_arg(
            '--examples-jar', dest='examples_jar',
            default=_EXAMPLES_JAR_URI,
            help=('URI of Hadoop example jar. Usually the default'
                  ' is fine, but on EMR, you should use '
                  ' file:///home/hadoop/hadoop-examples.jar'
                  ' with 3.x AMIs and earlier')
        )

    def steps(self):
        jar = self.options.examples_jar

        if self.options.use_main_class:
            jar_step = JarStep(
                jar=jar,
                args=[GENERIC_ARGS, INPUT, OUTPUT],
                main_class=_WORDCOUNT_MAIN_CLASS,
            )
        else:
            jar_step = JarStep(
                jar=jar,
                args=['wordcount', GENERIC_ARGS, INPUT, OUTPUT],
            )

        return [
            jar_step,
            MRStep(
                mapper=self.mapper,
                combiner=self.reducer,
                reducer=self.reducer,
            )
        ]

    def mapper(self, key, freq):
        yield int(freq), 1

    def reducer(self, freq, counts):
        yield freq, sum(counts)

    def pick_protocols(self, step_num, step_type):
        """Use RawProtocol to read output from the jar."""
        read, write = super(MRJarStepExample, self).pick_protocols(
            step_num, step_type)
        if (step_num, step_type) == (1, 'mapper'):
            read = RawProtocol().read

        return read, write


if __name__ == '__main__':
    MRJarStepExample.run()



# Copyright 2009-2016 Yelp and Contributors
# Copyright 2017 Yelp
# Copyright 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# see case.py for definition of mock_clusters and mock_gcs_fs
from copy import deepcopy

from google.api_core.exceptions import AlreadyExists
from google.api_core.exceptions import InvalidArgument
from google.api_core.exceptions import NotFound
from google.cloud.dataproc_v1beta2.types import Cluster
from google.cloud.dataproc_v1beta2.types import ClusterStatus
from google.cloud.dataproc_v1beta2.types import DiskConfig
from google.cloud.dataproc_v1beta2.types import Job
from google.cloud.dataproc_v1beta2.types import JobStatus

from mrjob.dataproc import _STATE_MATCHER_ACTIVE
from mrjob.dataproc import _cluster_state_name
from mrjob.dataproc import _job_state_name
from mrjob.dataproc import _zone_to_region
from mrjob.parse import is_uri
from mrjob.util import random_identifier


# account scopes that are included whether you ask for them or not
# for more info, see:
# https://cloud.google.com/dataproc/docs/reference/rest/v1/projects.regions.clusters#GceClusterConfig  # noqa
_MANDATORY_SCOPES = {
    'https://www.googleapis.com/auth/cloud.useraccounts.readonly',
    'https://www.googleapis.com/auth/devstorage.read_write',
    'https://www.googleapis.com/auth/logging.write',
}

# account scopes that are included if you don't specify any
_DEFAULT_SCOPES = {
    'https://www.googleapis.com/auth/bigquery',
    'https://www.googleapis.com/auth/bigtable.admin.table',
    'https://www.googleapis.com/auth/bigtable.data',
    'https://www.googleapis.com/auth/devstorage.full_control',
}

# default boot disk size set by the API
_DEFAULT_DISK_SIZE_GB = 500

# actual properties taken from Dataproc
_DEFAULT_CLUSTER_PROPERTIES = {
    'distcp:mapreduce.map.java.opts': '-Xmx2457m',
    'distcp:mapreduce.map.memory.mb': '3072',
    'distcp:mapreduce.reduce.java.opts': '-Xmx2457m',
    'distcp:mapreduce.reduce.memory.mb': '3072',
    'hdfs:dfs.namenode.handler.count': '20',
    'hdfs:dfs.namenode.service.handler.count': '10',
    'mapred-env:HADOOP_JOB_HISTORYSERVER_HEAPSIZE': '1000',
    'mapred:mapreduce.map.cpu.vcores': '1',
    'mapred:mapreduce.map.java.opts': '-Xmx2457m',
    'mapred:mapreduce.map.memory.mb': '3072',
    'mapred:mapreduce.reduce.cpu.vcores': '1',
    'mapred:mapreduce.reduce.java.opts': '-Xmx2457m',
    'mapred:mapreduce.reduce.memory.mb': '3072',
    'mapred:yarn.app.mapreduce.am.command-opts': '-Xmx2457m',
    'mapred:yarn.app.mapreduce.am.resource.cpu-vcores': '1',
    'mapred:yarn.app.mapreduce.am.resource.mb': '3072',
    'spark-env:SPARK_DAEMON_MEMORY': '1000m',
    'spark:spark.driver.maxResultSize': '480m',
    'spark:spark.driver.memory': '960m',
    'spark:spark.executor.cores': '1',
    'spark:spark.executor.memory': '1152m',
    'spark:spark.yarn.am.memory': '1152m',
    'spark:spark.yarn.am.memoryOverhead': '384',
    'spark:spark.yarn.executor.memoryOverhead': '384',
    'yarn-env:YARN_TIMELINESERVER_HEAPSIZE': '1000',
    'yarn:yarn.nodemanager.resource.memory-mb': '3072',
    'yarn:yarn.scheduler.maximum-allocation-mb': '3072',
    'yarn:yarn.scheduler.minimum-allocation-mb': '256',
}


# convert strings (e.g. 'RUNNING') to enum values

def _cluster_state_value(state_name):
    return ClusterStatus.State.Value(state_name)


def _job_state_value(state_name):
    return JobStatus.State.Value(state_name)


class MockGoogleDataprocClient(object):
    """Common code for mock dataproc clients."""

    def __init__(
            self, mock_clusters, mock_jobs, mock_gcs_fs, mock_jobs_succeed,
            channel=None, credentials=None):
        self.channel = channel
        self.credentials = credentials

        # maps (project_id, region, cluster_name) to a
        # google.cloud.dataproc_v1beta2.types.Cluster
        self.mock_clusters = mock_clusters

        # maps (project_id, region, cluster_name, job_name) to a
        # google.cloud.dataproc_v1beta2.types.Job
        self.mock_jobs = mock_jobs

        # if False, mock jobs end in ERROR
        self.mock_jobs_succeed = mock_jobs_succeed

        # see case.py
        self.mock_gcs_fs = mock_gcs_fs

    def _expected_region(self):
        if self.channel:
            target = self.channel._channel.target()
            name = target.split('.')[0]
            if '-' in name:
                return '-'.join(name.split('-')[:-1])

        return 'global'

    def _check_region_matches_endpoint(self, region):
        expected = self._expected_region()

        if region != expected:
            raise InvalidArgument(
                "Region '%s' invalid or not supported by this endpoint;"
                " permitted regions: '[%s]'" % (region, expected))


class MockGoogleDataprocClusterClient(MockGoogleDataprocClient):

    """Mock out google.cloud.dataproc_v1beta2.ClusterControllerClient"""

    def create_cluster(self, project_id, region, cluster):
        self._check_region_matches_endpoint(region)

        # convert dict to object
        if not isinstance(cluster, Cluster):
            cluster = Cluster(**cluster)

        if cluster.project_id:
            if cluster.project_id != project_id:
                raise InvalidArgument(
                    'If provided, CreateClusterRequest.cluster.project_id must'
                    ' match CreateClusterRequest.project_id')
        else:
            cluster.project_id = project_id

        if not cluster.cluster_name:
            raise InvalidArgument('Cluster name is required')

        # add in default disk config
        for x in ('master', 'worker', 'secondary_worker'):
            field = x + '_config'
            conf = getattr(cluster.config, field, None)
            if conf and str(conf):  # empty DiskConfigs are still true-ish
                if not conf.disk_config:
                    conf.disk_config = DiskConfig()
                if not conf.disk_config.boot_disk_size_gb:
                    conf.disk_config.boot_disk_size_gb = _DEFAULT_DISK_SIZE_GB

        # update gce_cluster_config
        gce_config = cluster.config.gce_cluster_config

        # check region and zone_uri
        if region == 'global':
            if gce_config.zone_uri:
                cluster_region = _zone_to_region(gce_config.zone_uri)
            else:
                raise InvalidArgument(
                    "Must specify a zone in GCE configuration"
                    " when using 'regions/global'")
        else:
            cluster_region = region

        # add in default scopes and sort
        scopes = set(gce_config.service_account_scopes)

        if not scopes:
            scopes.update(_DEFAULT_SCOPES)
        scopes.update(_MANDATORY_SCOPES)

        gce_config.service_account_scopes[:] = sorted(scopes)

        # handle network_uri and subnetwork_uri
        if gce_config.network_uri and gce_config.subnetwork_uri:
            raise InvalidArgument('GceClusterConfiguration cannot contain both'
                                  ' Network URI and Subnetwork URI')

        if not (gce_config.network_uri or gce_config.subnetwork_uri):
            gce_config.network_uri = 'default'

        if gce_config.network_uri:
            gce_config.network_uri = _fully_qualify_network_uri(
                gce_config.network_uri, project_id)

        if gce_config.subnetwork_uri:
            gce_config.subnetwork_uri = _fully_qualify_subnetwork_uri(
                gce_config.subnetwork_uri, project_id, cluster_region)

        # add in default cluster properties
        props = cluster.config.software_config.properties

        for k, v in _DEFAULT_CLUSTER_PROPERTIES.items():
            if k not in props:
                props[k] = v

        # initialize cluster status
        cluster.status.state = _cluster_state_value('CREATING')

        cluster_key = (project_id, region, cluster.cluster_name)

        if cluster_key in self.mock_clusters:
            raise AlreadyExists('Already exists: Cluster ' +
                                _cluster_path(*cluster_key))

        self.mock_clusters[cluster_key] = cluster

    def delete_cluster(self, project_id, region, cluster_name):
        self._check_region_matches_endpoint(region)

        cluster_key = (project_id, region, cluster_name)

        cluster = self.mock_clusters.get(cluster_key)

        if not cluster:
            raise NotFound('Not found: Cluster ' + _cluster_path(*cluster_key))

        cluster.status.state = _cluster_state_value('DELETING')

    def get_cluster(self, project_id, region, cluster_name):
        self._check_region_matches_endpoint(region)

        cluster_key = (project_id, region, cluster_name)
        if cluster_key not in self.mock_clusters:
            raise NotFound(
                'Not Found: Cluster ' + _cluster_path(*cluster_key))

        cluster = self.mock_clusters[cluster_key]

        result = deepcopy(cluster)
        self._simulate_progress(project_id, region, cluster_name)
        return result

    def _simulate_progress(self, project_id, region, cluster_name):
        cluster_key = (project_id, region, cluster_name)
        cluster = self.mock_clusters[cluster_key]

        state_name = _cluster_state_name(cluster.status.state)

        if state_name == 'DELETING':
            del self.mock_clusters[cluster_key]
        else:
            # just move from STARTING to RUNNING
            cluster.status.state = _cluster_state_value('RUNNING')


class MockGoogleDataprocJobClient(MockGoogleDataprocClient):

    """Mock out google.cloud.dataproc_v1beta2.JobControllerClient"""

    def submit_job(self, project_id, region, job):
        self._check_region_matches_endpoint(region)

        # convert dict to object
        if not isinstance(job, Job):
            job = Job(**job)

        if not (project_id and job.reference.job_id):
            raise NotImplementedError('generation of job IDs not implemented')
        job_id = job.reference.job_id

        if not job.placement.cluster_name:
            raise InvalidArgument('Cluster name is required')

        # cluster must exist
        cluster_key = (project_id, region, job.placement.cluster_name)
        if cluster_key not in self.mock_clusters:
            raise NotFound('Not Found: Cluster ' + _cluster_path(*cluster_key))

        if not job.hadoop_job:
            raise NotImplementedError('only hadoop jobs are supported')

        if job.reference.project_id:
            if job.reference.project_id != project_id:
                raise InvalidArgument(
                    'If provided, SubmitJobRequest.job.job_reference'
                    '.project_id must match SubmitJobRequest.project_id')
        else:
            job.reference.project_id = project_id

        job.status.state = _job_state_value('SETUP_DONE')

        job_key = (project_id, region, job_id)

        if job_key in self.mock_jobs:
            raise AlreadyExists(
                'Already exists: Job ' + _job_path(*job_key))

        self.mock_jobs[job_key] = job

        return deepcopy(job)

    def get_job(self, project_id, region, job_id):
        self._check_region_matches_endpoint(region)

        job_key = (project_id, region, job_id)

        job = self.mock_jobs.get(job_key)

        if not job:
            raise NotFound('Not found: Job ' + _job_path(*job_key))

        result = deepcopy(job)
        self._simulate_progress(job)
        return result

    def list_jobs(self, project_id, region, page_size=None,
                  cluster_name=None, job_state_matcher=None):
        self._check_region_matches_endpoint(region)

        if page_size:
            raise NotImplementedError('page_size is not mocked')

        for job_key, job in self.mock_jobs.items():
            job_project_id, job_region, job_id = job_key

            if job_project_id != project_id:
                continue

            if job_region != region:
                continue

            if cluster_name and job.placement.cluster_name != cluster_name:
                continue

            if job_state_matcher:
                if job_state_matcher != _STATE_MATCHER_ACTIVE:
                    raise NotImplementedError(
                        'only ACTIVE job state matcher is mocked')

                if (_job_state_name(job.status.state) not in
                        ('PENDING', 'RUNNING', 'CANCEL_PENDING')):
                    continue

            yield deepcopy(job)

    def _simulate_progress(self, mock_job):
        state = _job_state_name(mock_job.status.state)

        if state == 'SETUP_DONE':
            mock_job.status.state = _job_state_value('PENDING')
        elif state == 'PENDING':
            mock_job.status.state = _job_state_value('RUNNING')
            # for now now, we just need this to be set
            mock_job.driver_output_resource_uri = (
                'gs://mock-bucket-%s/google-cloud-dataproc-metainfo/'
                'mock-cluster-id-%s/jobs/mock-job-%s/driveroutput' % (
                    random_identifier(), random_identifier(),
                    random_identifier()))
        elif state == 'RUNNING':
            if self.mock_jobs_succeed:
                mock_job.status.state = _job_state_value('DONE')
            else:
                mock_job.status.state = _job_state_value('ERROR')


def _cluster_path(project_id, region, cluster_name):
    return 'projects/%s/regions/%s/clusters/%s' % (
        project_id, region, cluster_name)


def _job_path(project_id, region, job_id):
    return 'projects/%s/regions/%s/jobs/%s'


def _fully_qualify_network_uri(uri, project_id):
    if '/' not in uri:  # just a name
        uri = 'projects/%s/global/networks/%s' % (project_id, uri)

    if not is_uri(uri):
        uri = 'https://www.googleapis.com/compute/v1/' + uri

    return uri


def _fully_qualify_subnetwork_uri(uri, project_id, region):
    if '/' not in uri:  # just a name
        uri = 'projects/%s/%s/subnetworks/%s' % (project_id, region, uri)

    if not is_uri(uri):
        uri = 'https://www.googleapis.com/compute/v1/' + uri

    return uri



# -*- coding: utf-8 -*-
# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for creating custom AMIs."""
import logging
import re

log = logging.getLogger(__name__)

# used to match the name of AMIs that we want to use on EMR
_EMR_BASE_AMI_NAME_RE = re.compile(r'^amzn-ami-hvm-[\d\.]*-x86_64-ebs$')


def describe_base_emr_images(ec2_client):
    """Fetch a list of Amazon Linux AMI images that are usable with EMR,
    with the most recent first. This can take several seconds.

    :param ec2_client: a boto3 EC2 client, which can be obtained from
                       :py:meth:`mrjob.emr.EMRJobRunner.make_ec2_client()`
                       or ``boto3.client('ec2')``

    For the sake of consistency, we have somewhat stricter requirements
    than `the AWS documentation <https://docs.aws.amazon.com/emr/latest/\
    ManagementGuide/emr-custom-ami.html#emr-custom-ami-considerations>`_.
    Specifically:

    * Amazon Linux (not Amazon Linux 2)
    * HVM virtualization
    * x86_64 architecture
    * single EBS volume
      * standard volume type (not GP2)
    * stable version (no "testing" or "rc", only numbers and dots)

    This only returns images going back to September 2016 (prior to that,
    EC2 used a different naming convention).

    This returns a dictionary for each image, in the same response format as
    `ec2_client.describe_images() <https://boto3.amazonaws.com/v1/\
    documentation/api/latest/reference/services/ec2.html#EC2.Client\
    .describe_images>`_. The
    *ImageId* field contains the AMI ID, and *Description* contains
    a human-readable description.
    """
    # DescribeImages' filtering is imperfect and slow, but this helps a bit
    images = ec2_client.describe_images(
        Owners=['amazon'],
        Filters=[
            dict(Name='architecture', Values=['x86_64']),
            dict(Name='root-device-type', Values=['ebs']),
            dict(Name='virtualization-type', Values=['hvm']),
        ],
    )['Images']

    # perform further filtering by name to pick out Amazon Linux
    images = [img for img in images
              if _EMR_BASE_AMI_NAME_RE.match(img.get('Name') or '')]

    # filter out any images that have multiple volumes
    # (this is implied by the naming convention, but just in case)
    images = [img for img in images
              if len(img.get('BlockDeviceMappings') or []) == 1]

    # require a CreationDate
    images = [img for img in images if img['CreationDate']]

    # put most recent images first
    images.sort(key=lambda img: img['CreationDate'], reverse=True)

    return images



# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mock boto3 EC2 support."""
from copy import deepcopy

from botocore.exceptions import ParamValidationError

from mrjob.aws import _DEFAULT_AWS_REGION

from .util import MockClientMeta


class MockEC2Client(object):
    """Mock out boto3 EC2 client

    :param mock_ec2_images: A list of image dictionaries to be returned
                            by :py:meth:`describe_images`
    """
    def __init__(self,
                 mock_ec2_images,
                 aws_access_key_id=None,
                 aws_secret_access_key=None,
                 aws_session_token=None,
                 endpoint_url=None,
                 region_name=None):

        self.mock_ec2_images = mock_ec2_images

        region_name = region_name or _DEFAULT_AWS_REGION
        if not endpoint_url:
            endpoint_url = 'https://ec2.%s.amazonaws.com' % region_name

        self.meta = MockClientMeta(
            endpoint_url=endpoint_url,
            region_name=region_name)

    def describe_images(self, Filters=None, Owners=None):
        images = []

        for image in self.mock_ec2_images:
            if not (Owners is None or image.get('ImageOwnerAlias') in Owners):
                continue

            if Filters and not _matches_image_filters(image, Filters):
                continue

            images.append(deepcopy(image))

        return dict(Images=images)


def _matches_image_filters(image, Filters):
    for Filter in Filters:
        if set(Filter) != {'Name', 'Values'}:
            raise ParamValidationError(
                report='Unknown parameter in Filters')

        field = _hyphen_to_camel(Filter['Name'])
        if not image.get(field) in Filter['Values']:
            return False

    return True


def _hyphen_to_camel(s):
    """Convert a string like ``root-device-type`` to ``RootDeviceType``"""
    return ''.join(part[0].upper() + part[1:] for part in s.split('-'))



# Copyright 2017 Yelp
# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test the diagnose script"""

from mrjob.tools.diagnose import main as diagnose_main

from tests.mock_boto3 import MockBoto3TestCase
from tests.py2 import patch


class StepPickingTestCase(MockBoto3TestCase):

    def setUp(self):
        super(StepPickingTestCase, self).setUp()

        self.pick_error = self.start(
            patch('mrjob.emr.EMRJobRunner._pick_error',
                  side_effect=StopIteration))

        self.log = self.start(
            patch('mrjob.tools.diagnose.log'))

        # don't print logging messages when we start the diagnose tool
        self.log_to_stream = self.start(
            patch('mrjob.launch.log_to_stream'))

    def test_single_failed_step(self):
        self.mock_emr_clusters['j-MOCKCLUSTER0'] = dict(
            _Steps=[
                dict(
                    Config=dict(Args=['python', 'mr_failure.py', '--mapper']),
                    Id='s-MOCKSTEP0',
                    Name='mr_failure',
                    Status=dict(State='FAILED'),
                )
            ]
        )

        self.assertRaises(StopIteration, diagnose_main, ['j-MOCKCLUSTER0'])

        self.pick_error.assert_called_once_with(
            dict(step_id='s-MOCKSTEP0'), 'streaming')

    def test_failed_step_followed_by_successful_step(self):
        self.mock_emr_clusters['j-MOCKCLUSTER0'] = dict(
            _Steps=[
                dict(
                    Config=dict(Args=['python', 'mr_failure.py', '--mapper']),
                    Id='s-MOCKSTEP0',
                    Name='mr_failure',
                    Status=dict(State='FAILED'),
                ),
                dict(
                    Config=dict(Args=['python', 'mr_success.py', '--mapper']),
                    Id='s-MOCKSTEP1',
                    Name='mr_success',
                    Status=dict(State='COMPLETED'),
                ),
            ]
        )

        self.assertRaises(StopIteration, diagnose_main, ['j-MOCKCLUSTER0'])

        self.pick_error.assert_called_once_with(
            dict(step_id='s-MOCKSTEP0'), 'streaming')

    def test_two_failed_steps(self):
        self.mock_emr_clusters['j-MOCKCLUSTER0'] = dict(
            _Steps=[
                dict(
                    Config=dict(Args=['python', 'mr_failure.py', '--mapper']),
                    Id='s-MOCKSTEP0',
                    Name='mr_failure',
                    Status=dict(State='FAILED'),
                ),
                dict(
                    Config=dict(Args=['python', 'mr_failure.py', '--mapper']),
                    Id='s-MOCKSTEP1',
                    Name='mr_failure',
                    Status=dict(State='FAILED'),
                ),
            ]
        )

        self.assertRaises(StopIteration, diagnose_main, ['j-MOCKCLUSTER0'])

        self.pick_error.assert_called_once_with(
            dict(step_id='s-MOCKSTEP1'), 'streaming')

    def test_no_failed_steps(self):
        self.mock_emr_clusters['j-MOCKCLUSTER0'] = dict(
            _Steps=[
                dict(
                    Config=dict(Args=['python', 'mr_success.py', '--mapper']),
                    Id='s-MOCKSTEP0',
                    Name='mr_success',
                    Status=dict(State='COMPLETED'),
                ),
            ]
        )

        self.assertRaises(SystemExit, diagnose_main, ['j-MOCKCLUSTER0'])

        self.assertTrue(self.log.error.called)
        self.assertFalse(self.pick_error.called)

    def test_failed_spark_step(self):
        self.mock_emr_clusters['j-MOCKCLUSTER0'] = dict(
            _Steps=[
                dict(
                    Config=dict(Args=[
                        'custom-spark-submit',
                        '--master',
                        'yarn',
                        '--deploy-mode',
                        'cluster',
                        'custom-spark.jar',
                        'more-args',
                    ]),
                    Id='s-MOCKSTEP0',
                    Name='mr_spark_jar',
                    Status=dict(State='FAILED'),
                ),
            ]
        )

        self.assertRaises(StopIteration, diagnose_main, ['j-MOCKCLUSTER0'])

        # doesn't distinguish between different kinds of spark steps
        self.pick_error.assert_called_once_with(
            dict(step_id='s-MOCKSTEP0'), 'spark')

    def test_pick_step_id(self):
        self.mock_emr_clusters['j-MOCKCLUSTER0'] = dict(
            _Steps=[
                dict(
                    Config=dict(Args=['python', 'mr_failure.py', '--mapper']),
                    Id='s-MOCKSTEP0',
                    Name='mr_failure',
                    Status=dict(State='FAILED'),
                ),
                dict(
                    Config=dict(Args=['python', 'mr_failure.py', '--mapper']),
                    Id='s-MOCKSTEP1',
                    Name='mr_failure',
                    Status=dict(State='FAILED'),
                ),
            ]
        )

        self.assertRaises(StopIteration, diagnose_main,
                          ['j-MOCKCLUSTER0', '--step-id', 's-MOCKSTEP0'])

        self.pick_error.assert_called_once_with(
            dict(step_id='s-MOCKSTEP0'), 'streaming')

    def test_pick_completed_step(self):
        self.mock_emr_clusters['j-MOCKCLUSTER0'] = dict(
            _Steps=[
                dict(
                    Config=dict(Args=['python', 'mr_failure.py', '--mapper']),
                    Id='s-MOCKSTEP0',
                    Name='mr_failure',
                    Status=dict(State='FAILED'),
                ),
                dict(
                    Config=dict(Args=['python', 'mr_success.py', '--mapper']),
                    Id='s-MOCKSTEP1',
                    Name='mr_success',
                    Status=dict(State='COMPLETED'),
                ),
            ]
        )

        self.assertRaises(StopIteration, diagnose_main,
                          ['j-MOCKCLUSTER0', '--step-id', 's-MOCKSTEP1'])

        self.pick_error.assert_called_once_with(
            dict(step_id='s-MOCKSTEP1'), 'streaming')

        self.assertTrue(self.log.warning.called)

    def test_bad_step_id(self):
        self.mock_emr_clusters['j-MOCKCLUSTER0'] = dict(
            _Steps=[
                dict(
                    Config=dict(Args=['python', 'mr_failure.py', '--mapper']),
                    Id='s-MOCKSTEP0',
                    Name='mr_failure',
                    Status=dict(State='FAILED'),
                )
            ]
        )

        # shouldn't trigger an error from the EMR API
        self.assertRaises(SystemExit, diagnose_main,
                          ['j-MOCKCLUSTER0', '--step-id', 's-ILLY'])

        self.assertTrue(self.log.error.called)
        self.assertFalse(self.pick_error.called)



# Copyright 2009-2017 Yelp and Contributors
# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test case to mock out boto3."""
import itertools
import os
import shutil
import tempfile
import time
from io import BytesIO

import boto3

from mrjob.aws import _boto3_now
from mrjob.emr import _EMR_LOG_DIR
from mrjob.emr import EMRJobRunner

from tests.mockssh import create_mock_ssh_script
from tests.mockssh import mock_ssh_dir
from tests.mr_two_step_job import MRTwoStepJob
from tests.py2 import MagicMock
from tests.py2 import patch
from tests.sandbox import SandboxedTestCase

from .ec2 import MockEC2Client
from .emr import MockEMRClient
from .iam import MockIAMClient
from .s3 import MockS3Client
from .s3 import MockS3Resource
from .s3 import add_mock_s3_data


class MockBoto3TestCase(SandboxedTestCase):

    # if a test needs to create an EMR client more than this many
    # times, there's probably a problem with simulating progress
    MAX_EMR_CLIENTS = 120

    def setUp(self):
        # patch boto3
        self.mock_ec2_images = []
        self.mock_emr_failures = set()
        self.mock_emr_self_termination = set()
        self.mock_emr_clusters = {}
        self.mock_emr_output = {}
        self.mock_iam_instance_profiles = {}
        self.mock_iam_role_attached_policies = {}
        self.mock_iam_roles = {}
        self.mock_s3_fs = {}

        self.emr_client = None  # used by simulate_emr_progress()
        self.emr_client_counter = itertools.repeat(
            None, self.MAX_EMR_CLIENTS)

        self.start(patch.object(boto3, 'client', self.client))
        self.start(patch.object(boto3, 'resource', self.resource))

        super(MockBoto3TestCase, self).setUp()

        # patch slow things
        self.mrjob_zip_path = None

        def fake_create_mrjob_zip(mocked_runner, *args, **kwargs):
            if not self.mrjob_zip_path:
                self.mrjob_zip_path = self.makefile('fake_mrjob.zip')

            mocked_runner._mrjob_zip_path = self.mrjob_zip_path

            return self.mrjob_zip_path

        self.start(patch.object(
            EMRJobRunner, '_create_mrjob_zip',
            fake_create_mrjob_zip))

        self.start(patch.object(time, 'sleep'))

    def add_mock_s3_data(self, data,
                         age=None, location=None, storage_class=None,
                         restore=None):
        """Update self.mock_s3_fs with a map from bucket name
        to key name to data."""
        add_mock_s3_data(
            self.mock_s3_fs, data, age, location, storage_class, restore)

    def add_mock_ec2_image(self, image):
        """Add information about a mock EC2 Image (AMI) to be returned by
        mock :py:meth:`~tests.mock_boto3.ec2.MockEC2Client.describe_images`.

        This will automatically fill `CreationDate`. Other fields you
        might want to fill include:

        * ``Architecture`` (e.g. ``'i386'``, ``'x86_64'``)
        * ``BlockDeviceMappings`` (e.g. ``[{'DeviceName': '/dev/sda1'}]``)
        * ``ImageOwnerAlias`` (e.g. ``'amazon'``, ``'aws-marketplace'``)
        * ``Name`` (e.g. ``amzn-ami-hvm-2017.09.1.20171120-x86_64-s3``)
        * ``RootDeviceType`` (e.g. ``'ebs'``, ``'instance-store'``)
        * ``VirtualizationType (e.g. ``'hvm'``, ``'paravirtual'``)
        """
        image = dict(image)

        # TODO: will eventually need to add a mock user ID to support
        # filtering by owner == 'self'

        if not image.get('CreationDate'):
            image['CreationDate'] = _boto3_now().strftime(
                '%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'

        self.mock_ec2_images.append(image)

    def add_mock_emr_cluster(self, cluster):
        if cluster['Id'] in self.mock_emr_clusters:
            raise ValueError('mock cluster %s already exists' % cluster.id)

        for field in ('_BootstrapActions', '_InstanceGroups', '_Steps'):
            cluster.setdefault(field, [])

        cluster.setdefault('Name', cluster['Id'][2:])
        cluster.setdefault('NormalizedInstanceHours', 0)

        self.mock_emr_clusters[cluster['Id']] = cluster

    def simulate_emr_progress(self, cluster_id):
        if not self.emr_client:
            self.emr_client = self.client('emr')

        self.emr_client._simulate_progress(cluster_id)

    def prepare_runner_for_ssh(self, runner, num_workers=0):
        # TODO: Refactor this abomination of a test harness

        # Set up environment variables
        os.environ['MOCK_SSH_VERIFY_KEY_FILE'] = 'true'

        # Create temporary directories and add them to MOCK_SSH_ROOTS
        master_ssh_root = tempfile.mkdtemp(prefix='master_ssh_root.')
        os.environ['MOCK_SSH_ROOTS'] = 'testmaster=%s' % master_ssh_root
        mock_ssh_dir('testmaster', _EMR_LOG_DIR + '/hadoop/history')

        if not hasattr(self, 'worker_ssh_roots'):
            self.worker_ssh_roots = []

        self.addCleanup(self.teardown_ssh, master_ssh_root)

        # Make the fake binary
        os.mkdir(os.path.join(master_ssh_root, 'bin'))
        self.ssh_bin = os.path.join(master_ssh_root, 'bin', 'ssh')
        create_mock_ssh_script(self.ssh_bin)

        # Make a fake keyfile so that the 'file exists' requirements are
        # satsified
        self.keyfile_path = os.path.join(master_ssh_root, 'key.pem')
        with open(self.keyfile_path, 'w') as f:
            f.write('I AM DEFINITELY AN SSH KEY FILE')

        # Tell the runner to use the fake binary
        runner._opts['ssh_bin'] = [self.ssh_bin]
        # Also pretend to have an SSH key pair file
        runner._opts['ec2_key_pair_file'] = self.keyfile_path

        # use fake hostname
        runner._address_of_master = MagicMock(return_value='testmaster')
        runner._master_private_ip = MagicMock(return_value='172.172.172.172')

        # re-initialize fs
        runner._fs = None

    # TODO: this should be replaced
    def add_worker(self):
        """Add a mocked worker to the cluster. Caller is responsible for
        setting runner._opts['num_ec2_instances'] to the correct number.
        """
        worker_num = len(self.worker_ssh_roots)
        new_dir = tempfile.mkdtemp(prefix='worker_%d_ssh_root.' % worker_num)
        self.worker_ssh_roots.append(new_dir)
        os.environ['MOCK_SSH_ROOTS'] += (':testmaster!testworker%d=%s'
                                         % (worker_num, new_dir))

    def teardown_ssh(self, master_ssh_root):
        shutil.rmtree(master_ssh_root)
        for path in self.worker_ssh_roots:
            shutil.rmtree(path)

    def make_runner(self, *args):
        """create a dummy job, and call make_runner() on it.
        Use this in a with block:

        with self.make_runner() as runner:
            ...
        """
        stdin = BytesIO(b'foo\nbar\n')
        mr_job = MRTwoStepJob(['-r', 'emr'] + list(args))
        mr_job.sandbox(stdin=stdin)

        return mr_job.make_runner()

    def launch(self, runner):
        """Launch a job but don't wait for it to finish."""
        runner._add_input_files_for_upload()
        runner._launch()

    def run_and_get_cluster(self, *args):
        # TODO: not sure why we include -v
        with self.make_runner('-v', *args) as runner:
            runner.run()
            return runner._describe_cluster()

    # mock boto3.client()
    def client(self, service_name, **kwargs):
        if service_name == 'ec2':
            kwargs['mock_ec2_images'] = self.mock_ec2_images
            return MockEC2Client(**kwargs)

        elif service_name == 'emr':
            try:
                next(self.emr_client_counter)
            except StopIteration:
                raise Exception(
                    'Too many connections to mock EMR, may be stalled')

            kwargs['mock_s3_fs'] = self.mock_s3_fs
            kwargs['mock_emr_clusters'] = self.mock_emr_clusters
            kwargs['mock_emr_failures'] = self.mock_emr_failures
            kwargs['mock_emr_self_termination'] = (
                self.mock_emr_self_termination)
            kwargs['mock_emr_output'] = self.mock_emr_output
            return MockEMRClient(**kwargs)

        elif service_name == 'iam':
            kwargs['mock_iam_instance_profiles'] = (
                self.mock_iam_instance_profiles)
            kwargs['mock_iam_roles'] = self.mock_iam_roles
            kwargs['mock_iam_role_attached_policies'] = (
                self.mock_iam_role_attached_policies)
            return MockIAMClient(**kwargs)

        elif service_name == 's3':
            kwargs['mock_s3_fs'] = self.mock_s3_fs
            return MockS3Client(**kwargs)
        else:
            raise NotImplementedError(
                'mock %s service not supported' % service_name)

    # mock boto3.resource()
    def resource(self, service_name, **kwargs):
        if service_name == 's3':
            kwargs['mock_s3_fs'] = self.mock_s3_fs
            return MockS3Resource(**kwargs)
        else:
            raise NotImplementedError(
                'mock %s resource not supported' % service_name)



# Copyright 2018 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from io import BytesIO
from os.path import join

from warcio.warcwriter import WARCWriter

from mrjob.examples.mr_phone_to_url import MRPhoneToURL

from tests.sandbox import SandboxedTestCase
from tests.job import run_job


class MRPhoneToURLTestCase(SandboxedTestCase):

    def test_empty(self):
        self.assertEqual(run_job(MRPhoneToURL()), {})

    def test_three_pages(self):
        wet1 = BytesIO()
        writer1 = WARCWriter(wet1, gzip=False)

        write_conversion_record(
            writer1, 'https://nophonenumbershere.info',
            b'THIS-IS-NOT-A-NUMBER')
        write_conversion_record(
            writer1, 'https://big.directory/',
            b'The Time: (612) 777-9311\nJenny: (201) 867-5309\n')

        wet2_gz_path = join(self.tmp_dir, 'wet2.warc.wet.gz')
        with open(wet2_gz_path, 'wb') as wet2:
            writer2 = WARCWriter(wet2, gzip=True)

            write_conversion_record(
                writer2, 'https://jseventplanning.biz/',
                b'contact us at +1 201 867 5309')

        self.assertEqual(
            run_job(MRPhoneToURL([wet2_gz_path, '-']),
                    raw_input=wet1.getvalue()),
            {'+12018675309': 'https://jseventplanning.biz/',
             '+16127779311': 'https://big.directory/'})


def write_conversion_record(writer, url, content):
    writer.write_record(writer.create_warc_record(
        url, 'conversion', warc_content_type='text/plain',
        payload=BytesIO(content)))



# Copyright 2019 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MRWordFreqCount, with a trivial combiner command.

Used to ensure that Spark runner can skip combiners that run commands.
"""
from mrjob.examples.mr_word_freq_count import MRWordFreqCount
from mrjob.job import MRJob


class MRWordFreqCountWithCombinerCmd(MRWordFreqCount):

    combiner = MRJob.combiner  # un-set combiner()

    def combiner_cmd(self):
        return 'cat'


if __name__ == '__main__':
    MRWordFreqCountWithCombinerCmd.run()
